# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from random import randint

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

from pyaoscx.pyaoscx_module import PyaoscxModule


class ACL(PyaoscxModule):
    """
    Provide configuration management for ACL on AOS-CX devices.
    """

    base_uri = "system/acls"
    resource_uri_name = "acls"

    indices = ["name", "list_type"]

    cfg_aces = ListDescriptor("cfg_aces")

    def __init__(self, session, name, list_type, uri=None, **kwargs):
        self.session = session
        # Assign IDs
        self.name = name
        self.list_type = list_type
        self._uri = uri
        # List used to determine attributes related to the ACL configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Use to manage ACL Entries
        self.cfg_aces = []
        # Attribute used to know if object was changed recently
        self.__modified = False
        # Set an initial random version
        self._update_version()

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for an ACL table entry and fill
            the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}{2}{3}".format(
            ACL.base_uri,
            self.name,
            self.session.api.compound_index_separator,
            self.list_type,
        )
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Remove fields because they are not needed for the PUT request
        if "name" in data:
            data.pop("name")
        if "list_type" in data:
            data.pop("list_type")
        # Delete unwanted data
        if "cfg_aces" in data:
            data.pop("cfg_aces")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the ACL is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["name", "list_type"]
            )

        # Set original attributes
        self.__original_attributes = data

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        # Clean ACL Entries settings
        if self.cfg_aces == []:
            # Set ACL Entries if any
            # Adds ACL Entries to parent ACL already
            from pyaoscx.acl_entry import AclEntry

            AclEntry.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system ACLs, and create a dictionary
            containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing ACLs IDs as keys and a Acl objects as
            values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        acl_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a Acl object
            indices, acl = ACL.from_uri(session, uri)
            acl_dict[indices] = acl

        return acl_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing ACL table
            entry. Checks whether the ACL exists in the switch. Calls
            self.update() if ACL being updated. Calls self.create() if a new
            ACL is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing ACL table entry.
        :return modified: True if Object was modified and a PUT request
            was made.
        """
        # Variable returned
        modified = False

        acl_data = utils.get_attrs(self, self.config_attrs)

        # Compare dictionaries
        if acl_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            # The version should change every time the ACL (or any of
            # its entries) change so that it is written to hardware
            self._update_version()
            acl_data["cfg_version"] = self.cfg_version

            post_data = json.dumps(acl_data)

            uri = "{0}/{1}{2}{3}".format(
                ACL.base_uri,
                self.name,
                self.session.api.compound_index_separator,
                self.list_type,
            )
            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = acl_data
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new ACL table entry. Only returns if no
            exception is raised.
        :return modified: Boolean, True if entry was created.
        """
        acl_data = utils.get_attrs(self, self.config_attrs)
        acl_data["name"] = self.name
        acl_data["list_type"] = self.list_type

        post_data = json.dumps(acl_data)

        try:
            response = self.session.request(
                "POST", ACL.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()

        # Object was modified, as it was created
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete ACL table entry.
        """
        uri = "{0}/{1}{2}{3}".format(
            ACL.base_uri,
            self.name,
            self.session.api.compound_index_separator,
            self.list_type,
        )

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a Acl object given a response_data.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param response_data: The response must be a dictionary of the form:
            { "{name},{list_type}": URL }, with URL being of the form:
            "/rest/v10.04/system/acls/{name},{list_type}"
        :return: Acl object.
        """
        acl_arr = session.api.get_keys(response_data, ACL.resource_uri_name)
        list_type = acl_arr[1]
        name = acl_arr[0]

        return ACL(session, name, list_type)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a Acl object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI
        :return indices, acl: tuple containing both the indices and Acl object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(
            r"(.*)acls/(?P<index1>.+)[,./-](?P<index2>.+)"
        )
        name = index_pattern.match(uri).group("index1")
        list_type = index_pattern.match(uri).group("index2")

        # Create Acl object
        acl = ACL(session, name, list_type)
        indices = "{0},{1}".format(name, list_type)

        return indices, acl

    def __str__(self):
        return "ACL name:{0}, list_type:{1}".format(self.name, self.list_type)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific ACL URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}{3}{4}".format(
                self.session.resource_prefix,
                ACL.base_uri,
                self.name,
                self.session.api.compound_index_separator,
                self.list_type,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def _update_version(self):
        """
        Whenever the ACL (or any of its entries) change,the version should be
            updated so that it gets written to hardware. If the version doesn't
            change, the new configuration won't get to the hardware.
        """

        new_cfg_version = randint(-9007199254740991, 9007199254740991)

        if self.materialized:
            if hasattr(self, "cfg_version"):
                logging.warning(
                    "ACL %s was modified, but the version wasn't, "
                    "so the version was changed automatically to %d",
                    str(self),
                    new_cfg_version,
                )
            else:
                logging.warning(
                    "ACL %s didn't have a version configured. %d was added",
                    str(self),
                    new_cfg_version,
                )

        self.cfg_version = new_cfg_version

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def add_acl_entry(
        self,
        sequence_num,
        action,
        count=None,
        protocol=None,
        src_ip=None,
        dst_ip=None,
        dst_l4_port_min=None,
        dst_l4_port_max=None,
        src_mac=None,
        dst_mac=None,
        ethertype=None,
    ):
        """
        Create an AclEntry object, ACL Entry already exists, value passed
            won't update the entry.
        :param sequence_num: Integer number of the sequence
        :param action: Action should be either "permit" or "deny"
        :param count: Optional boolean flag that when true, will make entry
            increment hit count for matched packets
        :param protocol: Optional integer IP protocol number
        :param src_ip: Optional source IP address. Both IPv4 and IPv6 are
            supported.
            Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_ip: Optional destination IP address. Both IPv4 and IPv6
            are supported.
            Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_l4_port_min: Optional minimum L4 port number in range; used
            in conjunction with dst_l4_port_max.
        :param dst_l4_port_max: Optional maximum L4 port number in range; used
            in conjunction with dst_l4_port_min.
        :param src_mac: Optional source MAC address
            Example:
                '01:02:03:04:05:06'
        :param dst_mac: Optional destination MAC address
            Example:
                '01:02:03:04:05:06'
        :param ethertype: Optional integer EtherType number
        :return acl_entry: A AclEntry object
        """
        # Create ACL Entry
        acl_entry_obj = self.session.api.get_module(
            self.session,
            "AclEntry",
            sequence_num,
            parent_acl=self,
            action=action,
            count=count,
            protocol=protocol,
            src_ip=src_ip,
            dst_ip=dst_ip,
            dst_l4_port_min=dst_l4_port_min,
            dst_l4_port_max=dst_l4_port_max,
            src_mac=src_mac,
            dst_mac=dst_mac,
            ethertype=ethertype,
        )

        # Try to obtain data; if not, create
        try:
            acl_entry_obj.get()
        except GenericOperationError:
            # Create object inside switch
            acl_entry_obj.apply()

        return acl_entry_obj

    def modify_acl_entry(
        self,
        sequence_num,
        action,
        count=None,
        src_ip=None,
        dst_ip=None,
        dst_l4_port_min=None,
        dst_l4_port_max=None,
        src_mac=None,
        dst_mac=None,
        ethertype=None,
    ):
        """
        Modify an existing ACL Entry.
        :param sequence_num: Integer number of the sequence.
        :param action: Action should be either "permit" or "deny".
        :param count: Optional boolean flag that when true, will make entry
            increment hit count for matched packets.
        :param src_ip: Optional source IP address. Both IPv4 and IPv6
            are supported.
            Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_ip: Optional destination IP address. Both IPv4 and IPv6
            are supported.
            Example:
                10.10.12.11/255.255.255.255
                2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff
        :param dst_l4_port_min: Optional minimum L4 port number in range; used
            in conjunction with dst_l4_port_max.
        :param dst_l4_port_max: Optional maximum L4 port number in range; used
            in conjunction with dst_l4_port_min.
        :param src_mac: Optional source MAC address
            Example:
                '01:02:03:04:05:06'
        :param dst_mac: Optional destination MAC address
            Example:
                '01:02:03:04:05:06'
        :param ethertype: Optional integer EtherType number.
        :return acl_entry: A AclEntry object.
        """
        # Create ACL Entry
        acl_entry_obj = self.session.api.get_module(
            self.session, "AclEntry", sequence_num, parent_acl=self
        )
        # Get AclEntry object data
        acl_entry_obj.get()

        # Modify data
        acl_entry_obj.modify(
            action,
            count,
            src_ip,
            dst_ip,
            dst_l4_port_min,
            dst_l4_port_max,
            src_mac,
            dst_mac,
            ethertype,
        )

        return acl_entry_obj

    def delete_all_acl_entries(self):
        """
        Delete all ACL Entries within an ACL.
        :return: True if object was changed
        """
        # Verify ACL has the latest data
        self.get()

        # Delete all entries
        self.cfg_aces = []

        # ACL Entries deleted
        # Object modified
        return True



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import re

from urllib.parse import quote_plus, unquote_plus

from netaddr import EUI as MacAddress
from netaddr import mac_eui48

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule

from pyaoscx.mac import Mac


class StaticMac(Mac):
    """
    Provide configuration management for Static MAC on AOS-CX devices.
    """

    indices = ["mac_addr"]
    resource_uri_name = "static_macs"

    def __init__(self, session, mac_addr, parent_vlan, uri=None, **kwargs):

        """
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device
        :param mac_addr: String MAC address, or netaddr EUI object.
            Example:
                '01:02:03:04:05:06'
        :param parent_vlan: Vlan object to which this MAC belongs
        :param uri: Optional string containing the uri of the MAC object
        """
        super(StaticMac, self).__init__(
            session, "static", mac_addr, parent_vlan, uri=uri
        )
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        self.display_name = "static Mac"

    def _mac_path(self):
        """
        Get the path for internal purposes.
        """
        return "{0}/{1}".format(
            self.base_uri, quote_plus(str(self.mac_address))
        )

    def _set_configuration_items(self, data, selector):
        # Determines if the Static MAC is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", ["mac_addr"])

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Static MAC.
            Checks whether the Static MAC exists in the switch. Calls
            self.update() if Static MAC is being updated. Calls self.create()
            if a new Static MAC is being created.
        """
        if not self._parent_vlan.materialized:
            self._parent_vlan.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Static MAC.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Transform the MAC object to string to be sent to the switch
        self.mac_addr = str(self.mac_address)
        # Create the address representation for the URI
        static_mac_data = utils.get_attrs(self, self.config_attrs)

        if "port" in static_mac_data:
            static_mac_data["port"] = self.port.get_info_format()

        return self._put_data(static_mac_data)

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new Static MAC using the object's
            attributes as POST body. Only returns if no exception is raised.
        :return modified: Boolean, True if entry was created
        """
        self.mac_addr = str(self.mac_address)
        static_mac_data = utils.get_attrs(self, self.config_attrs)
        # Set attributes
        static_mac_data["mac_addr"] = self.mac_addr
        static_mac_data["vlan"] = self._parent_vlan.get_uri()
        if hasattr(self, "port") and self.port is not None:
            static_mac_data["port"] = self.port.get_uri()

        return self._post_data(static_mac_data)

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete Static MAC mac_addr from interface on
            the switch.
        """
        reference_address = quote_plus(str(self.mac_address))
        uri = "{0}/{1}".format(self.base_uri, reference_address)

        self._send_data(uri, None, "DELETE", "Delete")

        # Delete back reference from VRF
        for static_mac in self._parent_vlan.static_macs:
            if static_mac.mac_address == self.mac_address:
                self._parent_vlan.static_macs.remove(static_mac)

        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_vlan, response_data):
        """
        Create a Static MAC object given a response_data related to the IP6
            mac_addr object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: parent Vlan object where Static MAC is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                mac_addr: "/rest/v10.04/interface/static_macs/mac_addr"
            }
        :return: Static MAC object
        """
        mac_format = mac_eui48
        mac_format.word_sep = ":"
        static_mac_arr = session.api.get_keys(
            response_data, StaticMac.resource_uri_name
        )
        mac_addr = static_mac_arr[0]
        return StaticMac(session, mac_addr, parent_vlan)

    @classmethod
    def from_uri(cls, session, parent_vlan, uri):
        """
        Create a StaticMac object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_vlan: Parent VLAN class where Static MAC is stored.
        :param uri: a String with a URI.
        :return index, static_mac_obj: tuple containing both the StaticMac
            Object and the static_mac's mac_address.
        """
        mac_format = mac_eui48
        mac_format.word_sep = ":"
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)static_macs/(?P<index>.+)")
        reference_mac_addr = index_pattern.match(uri).group("index")

        mac_addr = MacAddress(
            unquote_plus(reference_mac_addr), dialect=mac_format
        )

        static_mac_obj = StaticMac(session, mac_addr, parent_vlan, uri=uri)

        return reference_mac_addr, static_mac_obj

    @PyaoscxModule.deprecated
    def get_uri(self):
        # TODO: remove this method in favor of uri_path once all
        # modules have been translated to the 'properties' style
        return self.uri_path

    @property
    def uri_path(self):
        """
        Method used to obtain the specific Static MAC URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = self.session.resource_prefix + self._mac_path()

        return self._uri

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class StaticNexthop(PyaoscxModule):
    """
    Provide configuration management for Static Nexthop settings on AOS-CX
        devices.
    """

    indices = ["id"]
    resource_uri_name = "static_nexthops"

    def __init__(self, session, _id, parent_static_route, uri=None, **kwargs):

        self.session = session
        # Assign IP
        self.id = _id
        # Assign parent StaticRoute object
        self.__set_static_route(parent_static_route)
        self._uri = uri
        # List used to determine attributes related to the Static Nexthop
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)

        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_static_route(self, parent_static_route):
        """
        Set parent StaticRoute object as an attribute for the StaticNexthop
            object.
        :param parent_static_route a Static_Route object.
        """
        # Set parent Static Route
        self.__parent_static_route = parent_static_route

        # Set URI
        self.base_uri = "{0}/{1}/static_nexthops".format(
            self.__parent_static_route.base_uri,
            self.__parent_static_route.reference_address,
        )

        # Verify Static Nexthop data doesn't exist already for Static Route
        for static_nexthop in self.__parent_static_route.static_nexthops:
            if static_nexthop.id == self.id:
                # Make list element point to current object
                static_nexthop = self
            else:
                # Add self to static_nexthops list in parent Static Route
                self.__parent_static_route.static_nexthops.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Static Nexthop table entry
            and fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.id)
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the Static Nexthop is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", ["id"])

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "id" in self.__original_attributes:
            self.__original_attributes.pop("id")

        # If the Static Route has a port inside the switch
        if "port" in data and self.port:
            port_response = self.port
            interface_cls = self.session.api.get_module(
                self.session, "Interface", ""
            )
            # Set port as a Interface Object
            self.port = interface_cls.from_response(
                self.session, port_response
            )
            # Materialize port
            self.port.get()

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session, parent_static_route):
        """
        Perform a GET call to retrieve all system Static Nexthop objects
            related to a Static Route, and create a dictionary containing
            StaticNexthop objects.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_static_route: StaticRoute object, parent for the Static
            Nexthops.
        :return: Dictionary containing Static Nexthop IDs as keys and a Static
            NexthopThis objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/static_nexthops".format(
            parent_static_route.base_uri, parent_static_route.reference_address
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        static_nexthop_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a StaticNexthop object and adds it to parent static_route
            # list
            _id, static_nexthop = StaticNexthop.from_uri(
                session, parent_static_route, uri
            )
            # Load all Static Nexthop data from within the Switch
            static_nexthop.get()
            static_nexthop_dict[_id] = static_nexthop

        return static_nexthop_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing Static Nexthop.
            Checks whether the static_nexthop exists in the switch. Calls
            self.update() if Static Nexthop being updated. Calls self.create()
            if a new Static Nexthop is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        if not self.__parent_static_route.materialized:
            self.__parent_static_route.apply()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing static_nexthop.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        static_nexthop_data = utils.get_attrs(self, self.config_attrs)

        # Get port uri
        if hasattr(self, "port") and self.port:
            static_nexthop_data["port"] = self.port.get_info_format()

        uri = "{0}/{1}".format(self.base_uri, self.id)

        # Compare dictionaries
        if static_nexthop_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:
            post_data = json.dumps(static_nexthop_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = static_nexthop_data

            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new static_nexthop. Only returns if no
            exception is raised.
        :return modified: Boolean, True if entry was created
        """
        static_nexthop_data = utils.get_attrs(self, self.config_attrs)
        static_nexthop_data["id"] = self.id

        # Get port uri
        if hasattr(self, "port") and self.port:
            static_nexthop_data["port"] = self.port.get_info_format()

        post_data = json.dumps(static_nexthop_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was created, thus modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete StaticNexthop table entry.
        """
        uri = "{0}/{1}".format(self.base_uri, self.id)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from VRF
        for static_nexthop in self.__parent_static_route.static_nexthops:
            if static_nexthop.id == self.id:
                self.__parent_static_route.static_nexthops.remove(
                    static_nexthop
                )

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_static_route, response_data):
        """
        Create a StaticNexthop object given a response_data related to the
            Static Nexthop ID object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_static_route: parent StaticRoute object where Static
            Nexthop is stored.
        :param response_data: The response must be a dictionary of the form:
            {
                id: "/rest/v10.04/system/static_routes/static_nexthops/id"
            }
        :return: StaticNexthop object.
        """
        static_nexthop_arr = session.api.get_keys(
            response_data, StaticNexthop.resource_uri_name
        )
        _id = static_nexthop_arr[0]
        return StaticNexthop(session, _id, parent_static_route)

    @classmethod
    def from_uri(cls, session, parent_static_route, uri):
        """
        Create a StaticNexthop object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_static_route: parent static_route class where
            static_nexthop is stored.
        :param uri: a String with a URI.
        :return index, static_nexthop: tuple containing both the Static Nexthop
            object and the Static Nexthop's ID.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)static_nexthops/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create static_nexthop object
        static_nexthop_obj = StaticNexthop(
            session, index, parent_static_route, uri=uri
        )

        return index, static_nexthop_obj

    def __str__(self):
        return "Static Nexthop: {0}".format(self.id)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific Static Nexthop URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_id,
                self.base_uri,
                self.id,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    @classmethod
    def get_next_id(cls, session, parent_static_route):
        """
        Method used to obtain the ID for the next Static Nexthop. Thus perform
            a GET call to retrieve all system Static Nexthop inside a Static
            Route, and with it determine the next ID.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_static_route: StaticRoute object, parent for the Static
            Nexthops.
        :return new_id: Integer with the new Id for the next Static Nexthop.
        """
        logging.info("Retrieving the switch static_nexthop")

        uri = "{0}/{1}/static_nexthops".format(
            parent_static_route.base_uri, parent_static_route.reference_address
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        next_id = None
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Obtain ID from uri
            _id, static_nexthop = StaticNexthop.from_uri(
                session, parent_static_route, uri
            )
            next_id = int(_id)

        # Set next ID
        if next_id is not None:
            next_id += 1
        else:
            next_id = 0
        return next_id

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from copy import deepcopy
from urllib.parse import quote_plus, unquote_plus
from warnings import warn

from netaddr import mac_eui48
from netaddr import EUI as MacAddress
from netaddr.core import AddrFormatError

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.parameter_error import ParameterError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils
from pyaoscx.utils.list_attributes import ListDescriptor

import pyaoscx.vrf as vrf_mod

from pyaoscx.ipv6 import Ipv6
from pyaoscx.vlan import Vlan

from pyaoscx.pyaoscx_module import PyaoscxModule


class Interface(PyaoscxModule):
    """
    Provide configuration management for Interface on AOS-CX devices.
    """

    base_uri = "system/interfaces"
    indices = ["name"]
    resource_uri_name = "interfaces"

    ip6_addresses = ListDescriptor("ip6_addresses")

    def __init__(self, session, name, uri=None, ip6_addresses=[], **kwargs):
        self.session = session
        self._uri = uri

        # List used to determine attributes related to the configuration
        self.config_attrs = []
        self.materialized = False

        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}

        # Set name, percents name and determine if Interface is a LAG
        self.name = ""
        self.__set_name(name)

        # List of previous interfaces before update
        # used to verify if an interface is deleted from lag
        self.__prev_interfaces = []

        # Set ip6 addresses
        self.ip6_addresses = ip6_addresses

        # Type required for configuration
        self.type = None
        # Set type
        self.__set_type()

        # Check if data should be added to object
        if self.__is_special_type:
            utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    @property
    def modified(self):
        return self.__modified

    def __set_name(self, name):
        """
        Set name attribute in the proper form for Interface object. Also sets
            the "percents name"-the name with any special characters replaced
            with percent-encodings.
        :param name: Interface name.
        """
        # Add attributes to class
        self.name = None
        self.percents_name = None

        if r"%2F" in name or r"%2C" in name or r"%3A" in name:
            self.name = unquote_plus(name)
            self.percents_name = name
        else:
            self.name = name
            self.percents_name = quote_plus(self.name)

    def __set_type(self):
        """
        Set Interface type when creating an Interface Object.
        """
        # Define all patterns
        lag_pattern = re.compile(r"lag[0-9]+$")
        loopback_pattern = re.compile(r"loopback[0-9]+$")
        tunnel_pattern = re.compile(r"tunnel(.*)")
        vlan_pattern = re.compile(r"vlan[0-9]+$")
        vxlan_pattern = re.compile(r"vxlan(.*)")

        # Sets interface as a special type
        self.__is_special_type = True

        if lag_pattern.match(self.name):
            self.type = "lag"
        elif loopback_pattern.match(self.name):
            self.type = "loopback"
        elif tunnel_pattern.match(self.name):
            self.type = "tunnel"
        elif vlan_pattern.match(self.name):
            self.type = "vlan"
        elif vxlan_pattern.match(self.name):
            self.type = "vxlan"
        else:
            self.__is_special_type = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a Interface table entry.
        :param depth: Integer deciding how many levels into the API JSON
            that references will be returned.
        :param selector: Alphanumeric option to select specific
            information to return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(Interface.base_uri, self.percents_name)

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the module is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(
                self, data, "config_attrs", ["name", "type"]
            )

        # Set original attributes
        self.__original_attributes = deepcopy(data)

        # Set a list of interfaces as an attribute
        if hasattr(self, "interfaces") and self.interfaces is not None:
            interfaces_list = []
            # Get all URI elements in the form of a list
            uri_list = self.session.api.get_uri_from_data(self.interfaces)

            for uri in uri_list:
                # Create an Interface object
                name, interface = Interface.from_uri(self.session, uri)

                # Check for circular reference
                # No need to get() if it's circular; it is already
                # materialized. Just set flag
                if name == self.name:
                    interface.materialized = True
                else:
                    # Materialize interface
                    interface.get()

                # Add interface to list
                interfaces_list.append(interface)

            # Set list as Interfaces
            self.interfaces = interfaces_list
            # Set list of previous Interfaces
            self.__prev_interfaces = list(self.interfaces)

        # Set VRF
        if hasattr(self, "vrf") and self.vrf is not None:
            # Set VRF as a Vrf object
            vrf_obj = vrf_mod.Vrf.from_response(self.session, self.vrf)
            self.vrf = vrf_obj
            # Materialized VRF
            self.vrf.get()

        # Set VLAN
        if hasattr(self, "vlan_tag") and self.vlan_tag is not None:
            # Set vlan_tag as a Vlan object
            vlan_obj = Vlan.from_response(self.session, self.vlan_tag)
            self.vlan_tag = vlan_obj
            # Materialized Vlan
            self.vlan_tag.get()

        # vlan_trunks
        # Set a list of VLANs as an attribute
        if hasattr(self, "vlan_trunks") and self.vlan_trunks is not None:
            vlan_trunks = []
            # Get all URI elements in the form of a list
            uri_list = self.session.api.get_uri_from_data(self.vlan_trunks)

            for uri in uri_list:
                # Create a Vlan object
                vlan_id, vlan = Vlan.from_uri(self.session, uri)
                # Materialize VLAN
                vlan.get()
                # Add VLAN to dictionary
                vlan_trunks.append(vlan)
            # Set list as VLANs
            self.vlan_trunks = vlan_trunks

        # Set all ACLs
        from pyaoscx.acl import ACL

        if hasattr(self, "aclmac_in_cfg") and self.aclmac_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclmac_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclmac_in_cfg = acl

        if hasattr(self, "aclmac_out_cfg") and self.aclmac_out_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclmac_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclmac_out_cfg = acl

        if hasattr(self, "aclv4_in_cfg") and self.aclv4_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_in_cfg = acl

        if hasattr(self, "aclv4_out_cfg") and self.aclv4_out_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_out_cfg = acl

        if hasattr(self, "aclv4_routed_in_cfg") and self.aclv4_routed_in_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_routed_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_routed_in_cfg = acl

        if hasattr(self, "aclv4_routed_out_cfg") and self.aclv4_routed_out_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv4_routed_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv4_routed_out_cfg = acl

        if hasattr(self, "aclv6_in_cfg") and self.aclv6_in_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_in_cfg = acl

        if hasattr(self, "aclv6_out_cfg") and self.aclv6_out_cfg is not None:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_out_cfg = acl

        if hasattr(self, "aclv6_routed_in_cfg") and self.aclv6_routed_in_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_routed_in_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_routed_in_cfg = acl

        if hasattr(self, "aclv6_routed_out_cfg") and self.aclv6_routed_out_cfg:
            # Create Acl object
            acl = ACL.from_response(self.session, self.aclv6_routed_out_cfg)
            # Materialize Acl object
            acl.get()
            self.aclv6_routed_out_cfg = acl

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True

        if self.ip6_addresses == []:
            # Set IPv6 addresses if any
            # Loads IPv6 objects already into the Interface
            Ipv6.get_all(self.session, self)
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system Interfaces and create.
        a dictionary containing each Interface as a Interface Object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing Interface's name as key and a Interface
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", Interface.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        interfaces_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create an Interface object
            name, interface = Interface.from_uri(session, uri)

            interfaces_dict[name] = interface

        return interfaces_dict

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create an Interface object given a response_data related to the
            Interface object.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            { "<interface_name>": URL }, with URL:
            "/rest/v10.04/system/interfaces/<interface_name>"
        :return: Interface object.
        """
        interfaces_id_arr = session.api.get_keys(
            response_data, Interface.resource_uri_name
        )
        interface_name = interfaces_id_arr[0]
        return session.api.get_module(session, "Interface", interface_name)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create an Interface object given a interface URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return name, interface_obj: tuple containing both the Interface's name
            and an Interface object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)/(?P<index>.+)")
        name_percents = index_pattern.match(uri).group("index")
        name = unquote_plus(name_percents)
        # Create Interface object
        interface_obj = session.api.get_module(
            session, "Interface", name, uri=uri
        )

        return name, interface_obj

    @classmethod
    def get_facts(cls, session):
        """
        Perform a GET call to retrieve all Interfaces and their data.
        :param cls: Class reference.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return facts: Dictionary containing Interface IDs as keys and
            Interface objects as values.
        """
        logging.info("Retrieving the switch interfaces facts")

        # Set depth
        interface_depth = session.api.default_facts_depth

        # Build URI
        uri = "{0}?depth={1}".format(Interface.base_uri, interface_depth)

        try:
            # Try to get facts via GET method
            response = session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Load into json format
        facts = json.loads(response.text)

        return facts

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create an Interface Object. Only returns if no
            exception is raised.
        :return True if entry was created inside Device.
        """
        interface_data = utils.get_attrs(self, self.config_attrs)

        interface_data["name"] = self.name
        # Set Type
        if self.type is not None:
            interface_data["type"] = self.type

        post_data = json.dumps(interface_data)

        try:
            response = self.session.request(
                "POST", Interface.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all objects data
        self.get()

        return True

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to update or create a Interface or Port table entry.
            Checks whether the Interface exists in the switch. Calls
            self.update() if Interface is being updated. Calls self.create() if
            a Interface table entry is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete Interface table entry.
        """
        if not self.__is_special_type:
            self.initialize_interface_entry()
        else:
            # Delete Interface via a DELETE REQUEST
            uri = "{0}/{1}".format(Interface.base_uri, self.name)

            try:
                response = self.session.request("DELETE", uri)

            except Exception as e:
                raise ResponseError("DELETE", e)

            if not utils._response_ok(response, "DELETE"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            # Clean LAG from interfaces
            # Delete interface references
            for interface in self.__prev_interfaces:
                # If interface name is not the same as the current one
                if interface.name != self.name and self.type == "lag":
                    interface.__delete_lag(self)

            # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing Interface or Port
            table entry.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        # Get interface PUT data depending on the configuration attributes
        # list
        iface_data = utils.get_attrs(self, self.config_attrs)

        # Check if VRF is inside the data related to interface
        if hasattr(self, "vrf") and self.vrf is not None:
            # Set VRF in the correct format for PUT
            iface_data["vrf"] = self.vrf.get_info_format()

        # Check if vlan_tag is inside the data related to interface
        if hasattr(self, "vlan_tag") and self.vlan_tag is not None:
            # Set VLAN in the correct format for PUT
            iface_data["vlan_tag"] = self.vlan_tag.get_info_format()

        # Set interfaces into correct form
        if hasattr(self, "interfaces") and self.interfaces is not None:
            formatted_interfaces = {}

            # Check for interfaces no longer in LAG
            if self.__is_special_type and self.type == "lag":
                for element in self.__prev_interfaces:
                    # If element was deleted from interfaces
                    if element not in self.interfaces:
                        # Delete element reference to current LAG
                        element.__delete_lag(self)

            # Set prev interfaces with current ones
            # Copies interfaces
            self.__prev_interfaces = list(self.interfaces)

            # Set interfaces into correct form
            for element in self.interfaces:
                # If element is the same as current, ignore
                if element.name == self.name and self.type == "lag":
                    pass
                else:
                    # Verify object is materialized
                    if not element.materialized:
                        raise VerificationError(
                            "Interface {0}".format(element.name),
                            "Object inside interfaces not materialized",
                        )
                    formated_element = element.get_info_format()
                    formatted_interfaces.update(formated_element)

                    if self.type == "lag":
                        # New element being added to LAG
                        element.__add_member_to_lag(self)

            # Set values in correct form
            iface_data["interfaces"] = formatted_interfaces

        # Set VLANs into correct form
        if "vlan_trunks" in iface_data:
            formated_vlans = {}
            # Set VLANs into correct form
            for element in self.vlan_trunks:
                # Verify object is materialized
                if not element.materialized:
                    raise VerificationError(
                        "Vlan {0}".format(element),
                        "Object inside vlan trunks not materialized",
                    )
                formated_element = element.get_info_format()
                formated_vlans.update(formated_element)

            # Set values in correct form
            iface_data["vlan_trunks"] = formated_vlans

        # Set all ACLs
        if "aclmac_in_cfg" in iface_data and self.aclmac_in_cfg:
            # Set values in correct form
            iface_data["aclmac_in_cfg"] = self.aclmac_in_cfg.get_info_format()

        if "aclmac_out_cfg" in iface_data and self.aclmac_out_cfg:
            # Set values in correct form
            iface_data[
                "aclmac_out_cfg"
            ] = self.aclmac_out_cfg.get_info_format()

        if "aclv4_in_cfg" in iface_data and self.aclv4_in_cfg:
            # Set values in correct form
            iface_data["aclv4_in_cfg"] = self.aclv4_in_cfg.get_info_format()

        if "aclv4_out_cfg" in iface_data and self.aclv4_out_cfg:
            # Set values in correct form
            iface_data["aclv4_out_cfg"] = self.aclv4_out_cfg.get_info_format()

        if "aclv4_routed_in_cfg" in iface_data and self.aclv4_routed_in_cfg:
            # Set values in correct form
            iface_data[
                "aclv4_routed_in_cfg"
            ] = self.aclv4_routed_in_cfg.get_info_format()

        if "aclv4_routed_out_cfg" in iface_data and self.aclv4_routed_out_cfg:
            # Set values in correct form
            iface_data[
                "aclv4_routed_out_cfg"
            ] = self.aclv4_routed_out_cfg.get_info_format()

        if "aclv6_in_cfg" in iface_data and self.aclv6_in_cfg:
            # Set values in correct form
            iface_data["aclv6_in_cfg"] = self.aclv6_in_cfg.get_info_format()

        if "aclv6_out_cfg" in iface_data and self.aclv6_out_cfg:
            # Set values in correct form
            iface_data["aclv6_out_cfg"] = self.aclv6_out_cfg.get_info_format()

        if "aclv6_routed_in_cfg" in iface_data and self.aclv6_routed_in_cfg:
            # Set values in correct form
            iface_data[
                "aclv6_routed_in_cfg"
            ] = self.aclv6_routed_in_cfg.get_info_format()

        if "aclv6_routed_out_cfg" in iface_data and self.aclv6_routed_out_cfg:
            # Set values in correct form
            iface_data[
                "aclv6_routed_out_cfg"
            ] = self.aclv6_routed_out_cfg.get_info_format()

        uri = "{0}/{1}".format(Interface.base_uri, self.percents_name)

        # Compare dictionaries
        if iface_data == self.__original_attributes:
            # Object was not modified
            modified = False
        else:

            put_data = json.dumps(iface_data)

            try:
                response = self.session.request("PUT", uri, data=put_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )
            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = deepcopy(iface_data)
            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def __add_member_to_lag(self, lag):
        """
        Perform PUT calls to configure a Port as a LAG member, and enable it.
        :param lag: pyaoscx.Interface object, to which the current port is
            being assigned to.
        """
        if not lag.materialized:
            raise VerificationError(
                "LAG {0}".format(lag.name),
                "Object is not materialized - Perform get()",
            )

        lag_name = lag.name
        # Extract LAG ID from LAG name
        lag_id = int(re.search("\\d+", lag_name).group())

        # Update Values
        try:
            self.user_config["admin"] = "down"
        except AttributeError:
            pass
        try:
            self.other_config["lacp-aggregation-key"] = lag_id
        except AttributeError:
            pass

        # Make a POST call and update values
        self.update()

    @PyaoscxModule.connected
    def __delete_lag(self, lag):
        """
        Perform PUT calls to update Interface, deleting the LAG reference
            inside of the Port that was assigned to that LAG.
        :param lag: pyaoscx.Interface object.
        """
        if not lag.materialized:
            raise VerificationError(
                "LAG {0}".format(lag.name),
                "Object is not materialized - Perform get()",
            )

        # Update Values
        try:
            self.user_config["admin"] = "down"
        except AttributeError:
            pass
        self.other_config.pop("lacp-aggregation-key", None)

        # Make a PUT call and update values
        self.update()

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific Interface URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                Interface.base_uri,
                self.percents_name,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    def __str__(self):
        """
        String containing the Interface name.
        :return: This class' string representation.
        """
        return "Interface Object, name: '{0}'".format(self.name)

    def __set_to_default(self):
        """
        Perform a PUT call to set Interface to default settings.
        :return: True if object was changed.
        """
        # Check for IPv6 addresses and delete them
        for address in self.ip6_addresses:
            address.delete()
        # Clean Attribute
        self.ip6_addresses = []

        interface_data = {}
        # Clear Interfaces
        if hasattr(self, "interfaces") and self.interfaces is not None:
            if self.__is_special_type and self.name == "lag":
                self.interfaces = []
                for element in self.__prev_interfaces:
                    # If element was deleted from interfaces
                    if element not in self.interfaces:
                        # Delete element reference to current LAG
                        try:
                            element.__delete_lag(self)
                        except AttributeError:
                            # Ignore error
                            pass
            else:
                self.interfaces = [self]

            # Set prev interfaces with current ones
            # Copies interfaces
            self.__prev_interfaces = list(self.interfaces)

        uri = "{0}/{1}".format(Interface.base_uri, self.percents_name)

        put_data = json.dumps(interface_data)

        try:
            response = self.session.request("PUT", uri, data=put_data)

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Setting %s default settings", self)

        # Update values with new ones
        self.get()
        return True

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    @PyaoscxModule.materialized
    def configure_l2(
        self,
        phys_ports=None,
        ipv4=None,
        vlan_ids_list=None,
        vlan_tag=1,
        lacp="passive",
        description=None,
        fallback_enabled=False,
        mc_lag=False,
        vlan_mode="native-untagged",
        trunk_allowed_all=False,
        native_vlan_tag=True,
    ):
        """
        Configure an Interface object, set the attributes to a L2 LAG and
            apply() changes inside Switch.
        :param phys_ports: List of physical ports to aggregate (e.g. ["1/1/1",
            "1/1/2", "1/1/3"]) or list of Interface Objects.
        :param ipv4: Optional list of IPv4 address to assign to the interface.
            If more than one is specified, all addresses except for the first
            are added as secondary_ip4. Defaults to nothing if not specified.
            Example: ['1.1.1.1', '2.2.2.2']
        :param vlan_ids_list: Optional list of integer VLAN IDs or VLAN objects
            to add as trunk VLANS. Defaults to empty list if not specified.
        :param vlan_tag: Optional VLAN ID or Vlan object to be added as
            vlan_tag. Defaults to VLAN 1.
        :param lacp: Should be either "passive" or "active." Defaults to
            "passive" if not specified.
        :param description: Optional description for the interface. Defaults
            to nothing if not specified.
        :param fallback_enabled: Boolean to determine if the LAG uses LACP
            fallback. Defaults to False if not specified.
        :param mc_lag: Boolean to determine if the LAG is multi-chassis.
            Defaults to False if not specified.
        :param vlan_mode: Vlan mode on Interface, should be access or trunk
            Defaults to 'native-untagged'.
        :param trunk_allowed_all: Flag for vlan trunk allowed all on L2
            interface, vlan_mode must be set to trunk.
        :param native_vlan_tag: Flag for accepting only tagged packets on
            VLAN trunk native, vlan_mode must be set to trunk.
        :return: True if object was changed.
        """
        # Set Physical Ports
        if phys_ports is not None:
            self.interfaces = []
            for port in phys_ports:
                port_obj = self.session.api.get_module(
                    self.session, "Interface", port
                )
                # Materialize Port
                port_obj.get()
                self.interfaces.append(port_obj)
        # Set lacp
        self.lacp = lacp

        # Set Mode, but keep it as it was if it receives None
        if vlan_mode:
            self.vlan_mode = vlan_mode

        if vlan_mode == "access":
            # Convert VLAN Tag into Object
            if isinstance(vlan_tag, int):
                # Create Vlan object
                vlan_tag = Vlan(self.session, vlan_tag)
                # Try to get data; if non-existent, throw error
                vlan_tag.get()
                self.vlan_tag = vlan_tag

        # Modify if trunk
        elif vlan_mode == "trunk":
            if vlan_tag is None:
                vlan_tag = 1

            # Create Vlan object
            vlan_tag = Vlan(self.session, vlan_tag)
            # Try to get data; if non-existent, throw error
            vlan_tag.get()
            # Set VLAN tag
            self.vlan_tag = vlan_tag

            # Set VLAN mode
            if native_vlan_tag:
                self.vlan_mode = "native-tagged"
            else:
                self.vlan_mode = "native-untagged"

            if not trunk_allowed_all:
                # Set VLAN Trunks
                if vlan_ids_list is not None:
                    self.vlan_trunks = []
                    for vlan in vlan_ids_list:
                        vlan_obj = Vlan(self.session, vlan)
                        vlan_obj.get()
                        self.vlan_trunks.append(vlan_obj)

        # Set description
        if description is not None:
            self.description = description

        # Set IPv4
        if ipv4 == []:
            self.ip4_address = None
            self.ip4_address_secondary = None
        elif isinstance(ipv4, list):
            self.ip4_address = ipv4[0]
            self.ip4_address_secondary = ipv4[1:]

        self.routing = False

        # Set all remaining attributes for a Lag to be an L2
        if self.__is_special_type:
            self.other_config["mclag_enabled"] = mc_lag
            self.other_config["lacp-fallback"] = fallback_enabled

        # Apply Changes inside Switch
        return self.apply()

    @PyaoscxModule.materialized
    def configure_l3(
        self,
        phys_ports=None,
        ipv4=None,
        ipv6=None,
        vrf="default",
        lacp="passive",
        description=None,
        fallback_enabled=False,
        mc_lag=False,
    ):
        """
        Configure an Interface object, if not materialized, materialize it and
            then set the attributes to a L3 LAG and apply() changes inside
            Switch.
        :param phys_ports: List of physical ports to aggregate (e.g. ["1/1/1",
            "1/1/2", "1/1/3"]) or list of Interface Objects.
        :param ipv4: Optional list of IPv4 address to assign
            to the interface. If more than one is specified, all addresses
            except for the first are added as secondary_ip4. Defaults to
            nothing if not specified. Example: ['1.1.1.1', '2.2.2.2']
        :param ipv6: String list of IPv6 addresses to assign to the interface.
            Defaults to nothing if not specified. List of A Ipv6 objects is
            accepted. Example:
            ['2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff']
        :param vrf: VRF to attach the SVI to. Defaults to "default" if not
            specified. A Vrf object is also accepted.
        :param lacp: Should be either "passive" or "active." Defaults to
            "passive" if not specified.
        :param description: Optional description for the interface. Defaults to
            nothing if not specified.
        :param fallback_enabled: Boolean to determine if the LAG uses LACP
            fallback. Defaults to False if not specified.
        :param mc_lag: Boolean to determine if the LAG is multi-chassis.
            Defaults to False if not specified.
        :return: True if object was changed.
        """
        # Set Physical Ports
        if phys_ports is not None:
            self.interfaces = []
            for port in phys_ports:
                port_obj = self.session.api.get_module(
                    self.session, "Interface", port
                )
                # Materialize Port
                port_obj.get()
                self.interfaces.append(port_obj)

        # Set IPv4
        if ipv4 == []:
            self.ip4_address = None
            self.ip4_address_secondary = None
        elif isinstance(ipv4, list):
            self.ip4_address = ipv4[0]
            self.ip4_address_secondary = ipv4[1:]

        # Set IPv6
        if ipv6 is not None and ipv6 != []:
            for ip_address in ipv6:
                # Verify if incoming address is a string
                if isinstance(ip_address, str):
                    # Create Ipv6 object -- add it to ipv6_addresses internal
                    # list
                    ip_address = self.session.api.get_module(
                        self.session,
                        "Ipv6",
                        ip_address,
                        parent_int=self,
                        type="global-unicast",
                        preferred_lifetime=604800,
                        valid_lifetime=2592000,
                        node_address=True,
                        ra_prefix=True,
                    )
                    # Try to get data, if non existent create
                    try:
                        # Try to obtain IPv6 address data
                        ip_address.get()
                    # If Ipv6 Object is non existent, create it
                    except GenericOperationError:
                        # Create IPv6 inside switch
                        ip_address.apply()
        # If IPv6 is empty, delete
        elif ipv6 == []:
            self.ip6_addresses = []

        # Set lacp
        self.lacp = lacp
        # Set description
        if description is not None:
            self.description = description

        # Set VRF
        if isinstance(vrf, str):
            vrf = vrf_mod.Vrf(self.session, vrf)
            vrf.get()
        self.vrf = vrf

        self.routing = True

        # Set all remaining attributes for a Lag to be an L3
        if self.__is_special_type:
            self.other_config["mclag_enabled"] = mc_lag
            self.other_config["lacp-fallback"] = fallback_enabled
        self.vlan_mode = "native-untagged"

        # Apply Changes inside Switch
        return self.apply()

    def configure_svi(
        self,
        vlan=None,
        ipv4=None,
        ipv6=None,
        vrf=None,
        description=None,
        int_type="vlan",
        user_config="up",
    ):
        """
        Configure a Interface table entry for a VLAN.
        :param vlan: Numeric ID of VLAN
            A Vlan object is also accepted
        :param ipv4: Optional list of IPv4 address to assign to the interface.
            If more than one is specified, all addresses except for the first
            are added as secondary_ip4. Defaults to nothing if not specified.
            Example: ['1.1.1.1'].
        :param ipv6: String list of IPv6 addresses to assign to the interface.
            Defaults to nothing if not specified. A Ipv6 object is also
            accepted. Example:
            ['2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff']
        :param vrf: VRF to attach the SVI to. Defaults to "default" if not
            specified. A Vrf object is also accepted.
        :param description: Optional description for the interface. Defaults
            to nothing if not specified.
        :param int_type: Type of interface; generally should be "vlan" for
            SVI's. Defaults to vlan.
        :return: True if object was changed.
        """
        if not self.materialized:
            raise VerificationError(
                "Interface {0}".format(self.name), "Object not materialized"
            )

        if vlan is not None:
            vlan_tag = vlan
            # Set VLAN Tag into Object
            if isinstance(vlan, int):
                name = "VLAN {0}".format(str(vlan))
                # Create Vlan object
                vlan_tag = self.session.api.get_module(
                    self.session, "Vlan", vlan, name=name
                )
                # Try to obtain data; if not, create
                try:
                    vlan_tag.get()
                except GenericOperationError:
                    # Create object inside switch
                    vlan_tag.apply()

            self.vlan_tag = vlan_tag

        # Set IPv4
        if ipv4 is not None and ipv4 != []:
            for i in range(len(ipv4)):
                if i == 0:
                    self.ip4_address = ipv4[i]
                else:
                    self.ip4_address_secondary.append(ipv4[i])
        # If IPv4 is empty, delete
        elif ipv4 == []:
            self.ip4_address = None
            self.ip4_address_secondary = None
        # Set IPv6
        if ipv6 is not None and ipv6 != []:
            for ip_address in ipv6:
                # Verify if incoming address is a string
                if isinstance(ip_address, str):
                    # Create Ipv6 object -- add it to ipv6_addresses internal
                    # list
                    ip_address = self.session.api.get_module(
                        self.session,
                        "Ipv6",
                        ip_address,
                        parent_int=self,
                        type="global-unicast",
                        preferred_lifetime=604800,
                        valid_lifetime=2592000,
                        node_address=True,
                        ra_prefix=True,
                    )
                    # Try to get data, if non existent create
                    try:
                        # Try to obtain IPv6 address data
                        ip_address.get()
                    # If Ipv6 Object is non existent, create it
                    except GenericOperationError:
                        # Create IPv6 inside switch
                        ip_address.apply()
        # If IPv6 is empty, delete
        elif ipv6 == []:
            self.ip6_addresses = []

        # Set VRF
        if vrf is not None:
            if isinstance(vrf, str):
                vrf = self.session.api.get_module(self.session, "Vrf", vrf)
                vrf.get()

            self.vrf = vrf

        # Set type
        self.type = int_type

        if description is not None:
            self.description = description

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def add_ipv4_address(self, ip_address):
        """
        Configure a Interface object to add a new IPv4 address to it and
            calls apply(), applying changes inside Switch.
        :param ip_address: IPv4 address to assign to the interface.
            Example: "1.1.1.1"
        :return: True if object was changed.
        """
        # Set incoming IPv4 address
        self.ip4_address = ip_address

        # Apply changes inside switch
        return self.apply()

    @PyaoscxModule.materialized
    def add_ipv6_address(self, ip_address, address_type="global-unicast"):
        """
        Configure a Interface object to append a IPv6 address to its
            ip6_addresses list and apply changes.
        :param ip_address: IPv6 address to assign to the interface.
            A Ipv6 object is also accepted. Example:
            '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'
        :param address_type: Type of Address. Defaults to global-unicast.
        :return: Ipv6 object.
        """
        # Verify if incoming address is a string
        if isinstance(ip_address, str):
            # Create Ipv6 object -- add it to ipv6_addresses internal list
            ipv6 = self.session.api.get_module(
                self.session,
                "Ipv6",
                ip_address,
                parent_int=self,
                type=address_type,
                preferred_lifetime=604800,
                valid_lifetime=2592000,
                node_address=True,
                ra_prefix=True,
            )
            # Try to get data, if non existent create
            try:
                # Try to obtain IPv6 address data
                ipv6.get()
            # If Ipv6 Object is non existent, create it
            except GenericOperationError:
                # Create IPv6 inside switch
                ipv6.apply()

        # Apply changes inside switch
        self.apply()

        return ipv6

    def delete_ipv6_address(self, ip_address):
        """
        Given a IPv6 address, delete that address from the current Interface.
        :param ip_address: IPv6 address to assign to the interface. A Ipv6
            object is also accepted. Example:
            '2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'.
        """
        if not self.materialized:
            raise VerificationError(
                "Interface {0}".format(self.name), "Object not materialized"
            )

        # Verify if incoming address is a object
        if isinstance(ip_address, Ipv6):
            # Obtain address
            ip_address = ip_address.address

        # Iterate through every address inside interface
        for add_obj in self.ip6_addresses:
            if add_obj.address == ip_address:
                # Removing address does an internal delete
                self.ip6_addresses.remove(add_obj)

    def configure_loopback(self, vrf, ipv4=None, description=None):
        """
        Configure an Interface object to create a Loopback Interface for a
            logical L3 Interface. If the Loopback Interface already exists and
            an IPv4 address is given, this function will update the IPv4
            address.
        :param vrf: VRF to attach the Loopback to. Defaults to "default"
            if not specified.
        :param ipv4: IPv4 address to assign to the interface. Defaults to
            nothing if not specified. Example: '1.1.1.1'
        :param description: Optional description for the interface. Defaults
            to nothing if not specified..
        :return: True if object was changed.
        """
        if not self.materialized:
            raise VerificationError(
                "Interface {0}".format(self.name), "Object not materialized"
            )

        # Set VRF
        if vrf is not None:
            if isinstance(vrf, str):
                vrf = self.session.api.get_module(self.session, "Vrf", vrf)
                vrf.get()

            self.vrf = vrf

        # Set IPv4
        if ipv4 is not None and ipv4 != []:
            for i in range(len(ipv4)):
                if i == 0:
                    self.ip4_address = ipv4[i]
                else:
                    self.ip4_address_secondary.append(ipv4[i])
        # If IPv4 is empty, delete
        elif ipv4 == []:
            self.ip4_address = None
            self.ip4_address_secondary = None

        if description is not None:
            self.description = description

        # Set all remaining attributes to create a loopback

        # when configuring a loopback interface, it must be powered on
        self.admin_state = "up"

        self.ospf_if_type = "ospf_iftype_loopback"

        # Apply changes to switch
        return self.apply()

    @PyaoscxModule.materialized
    def configure_vxlan(
        self, source_ipv4=None, description=None, dest_udp_port=4789
    ):
        """
        Configure VXLAN table entry for a logical L3 Interface. If the VXLAN
            Interface already exists and an IPv4 address is given, the function
            will update the IPv4 address.
        :param source_ipv4: Optional source IPv4 address to assign to the VXLAN
            interface. Defaults to nothing if not specified. Example:
            '1.1.1.1'.
        :param description: Optional description for the interface. Defaults
            to nothing if not specified.
        :param dest_udp_port: Optional Destination UDP Port that the VXLAN
            will use. Default is set to 4789.
        :return: True if object was changed.
        """
        # Set Values
        self.options["local_ip"] = source_ipv4
        self.options["vxlan_dest_udp_port"] = str(dest_udp_port)

        self.type = "vxlan"
        # when configuring a vxlan interface, it must be powered on
        self.admin_state = "up"

        if description is not None:
            self.description = description

        # Apply changes
        return self.apply()

    def set_vlan_mode(self, vlan_mode):
        """
        Set an L2 interface's VLAN mode. The options are 'native-tagged',
            'native-untagged', or 'access'.
        :param vlan_mode: A string, either 'native-tagged', 'native-untagged',
            or 'access', specifying the desired VLAN mode.
        :return: True if object was changed.
        """
        if not self.materialized:
            raise VerificationError(
                "Interface {0}".format(self.name), "Object not materialized"
            )

        # Set Values
        self.vlan_mode = vlan_mode
        self.routing = False

        # Apply changes
        return self.apply()

    def set_untagged_vlan(self, vlan):
        """
        Set the untagged VLAN on an access port.
        :param vlan: Numeric ID of VLAN to set on access port. A Vlan object is
            also accepted.
        :return: True if object was changed.
        """
        if not self.materialized:
            raise VerificationError(
                "Interface {0}".format(self.name), "Object not materialized"
            )

        # Set Values
        self.vlan_mode = "access"

        vlan_tag = vlan
        # Set Vlan Tag into Object
        if isinstance(vlan, int):
            # Create Vlan object
            vlan_tag = self.session.api.get_module(self.session, "Vlan", vlan)
            # Try to get data; if non-existent, throw error
            vlan_tag.get()

        # Set Vlan Tag
        self.vlan_tag = vlan_tag

        self.routing = False

        # Apply changes
        return self.apply()

    def add_vlan_trunks(self, vlan_trunk_ids):
        """
        Add specified VLANs to a trunk port. By default, this will also set
            the port to have 'no routing' and if there is not a native VLAN,
            will set the native VLAN to VLAN 1.
        :param vlan_trunk_ids: Dictionary of VLANs to specify
            as allowed on the trunk port.  If empty, the interface
            will allow all VLANs on the trunk.
        :return: True if object was changed.
        """
        # Set vlan Trunks
        if vlan_trunk_ids is not None:
            self.vlan_trunks = []

            for vlan in vlan_trunk_ids:

                vlan_obj = self.session.api.get_module(
                    self.session, "Vlan", vlan
                )
                vlan_obj.get()

                self.vlan_trunks.append(vlan_obj)

        self.routing = False

        # Set other values in case of None
        if self.vlan_mode is not None:
            self.vlan_mode = "native-untagged"

        if self.vlan_tag is not None:
            vlan_tag_obj = self.session.api.get_module(self.session, "Vlan", 1)
            vlan_tag_obj.get()
            self.vlan_tag = vlan_tag_obj

        # Apply Changes
        return self.apply()

    def set_native_vlan(self, vlan, tagged=True):
        """
        Set a VLAN to be the native VLAN on the trunk. Also gives the option to
            set the VLAN as tagged.
        :param vlan: Numeric ID of VLAN to add to trunk port. A Vlan object is
            also accepted
        :param tagged: Boolean to determine if True, the native VLAN will be
            set as the tagged VLAN. If False, the VLAN will be set as the
            native untagged VLAN. Defaults to True.
        :return: True if object was changed.
        """
        if tagged:
            self.vlan_mode = "native-tagged"
        else:
            self.vlan_mode = "native-untagged"

        vlan_tag = vlan
        # Set Vlan Tag into Object
        if isinstance(vlan_tag, int):
            # Create Vlan object
            vlan_tag = self.session.api.get_module(self.session, "Vlan", vlan)
            # Try to get data; if non-existent, throw error
            vlan_tag.get()

        self.vlan_tag = vlan_tag

        self.routing = False

        # Flag used to check if the incoming vlan has to be added to trunks
        add = True
        # Verify native vlan is in vlan trunks
        for vlan_obj in self.vlan_trunks:
            # Check vlan
            if vlan_obj.id == vlan:
                # Don't add
                add = False
        if add:
            # Add new vlan to vlan trunks
            self.vlan_trunks.append(self.vlan_tag)

        # Apply Changes
        return self.apply()

    def delete_vlan(self, vlan):
        """
        Delete a VLAN from a trunk port.
        :param vlan: Numeric ID of VLAN to delete from the trunk port.
            A Vlan object is also accepted.
        :return: True if successfully deleted.
        """
        # Import VLAN to  identify object type
        from pyaoscx.vlan import Vlan

        if isinstance(vlan, Vlan):
            vlan_id = vlan.id
        else:
            vlan_id = vlan

        deleted = False
        # Iterate through vlan trunks in search of the vlan
        for vlan_obj in self.vlan_trunks:
            if vlan_obj.id == vlan_id:
                # Delete vlan from vlan trunks
                self.vlan_trunks.remove(vlan_obj)
                deleted = True
        # Apply Changes
        self.apply()

        return deleted

    def add_port_to_lag(self, interface):
        """
        Configure a Port as a LAG member, and also enable the port. Add port
            to list of interfaces inside Interface object.
        :param interface: Alphanumeric name of the interface. A Interface
            object is also accepted.
        :return: True if object was changed.
        """
        # Identify interface variable type
        if isinstance(interface, str):
            # Create Interface Object
            interface_obj = self.session.api.get_module(
                self.session, "Interface", interface
            )
            # Try to get data; if non-existent, throw error
            interface_obj.get()

        elif isinstance(interface, Interface):
            interface_obj = interface

        for member in self.interfaces:
            # Check existance inside members
            if member.name == interface_obj.name:
                # Stop execution
                return False

        # Add interface as a member of the lag
        self.interfaces.append(interface_obj)

        # Apply changes
        return self.apply()

    def remove_port_from_lag(self, interface):
        """
        Remove a Port from LAG, and also disable the port. Remove port from
            list of interfaces inside Interface object.
        :param interface: Alphanumeric name of the interface. A Interface
            object is also accepted
        :return: True if object was changed.
        """
        if not self.__is_special_type:
            raise VerificationError(
                "Interface {0}".format(self.name),
                "Interface object must be a lag to remove a Port",
            )

        # Identify interface type
        if isinstance(interface, Interface):
            interface_name = interface.name
        elif isinstance(interface, str):
            interface_name = interface

        for member in self.interfaces:
            # Check existence inside members
            if member.name == interface_name:
                # Remove interface from Member
                self.interfaces.remove(member)

        # When changes are applied, port is disabled and lacp key changed
        return self.apply()

    def clear_acl(self, acl_type):
        """
        Clear an interface's ACL.
        :param acl_type: Type of ACL: options are 'aclv4_out', 'aclv4_in',
            'aclv6_in', or 'aclv6_out'.
        :return: True if object was changed.
        """
        if acl_type == "ipv6":
            self.aclv6_in_cfg = None
            self.aclv6_in_cfg_version = None
        if acl_type == "ipv4":
            self.aclv4_in_cfg = None
            self.aclv4_in_cfg_version = None
        if acl_type == "mac":
            self.aclmac_in_cfg = None
            self.aclmac_in_cfg_version = None

        # Apply Changes
        return self.apply()

    def initialize_interface_entry(self):
        """
        Initialize Interface to its default state.
        :return: True if object was changed.
        """
        # Set interface to default settings
        return self.__set_to_default()

    @property
    def admin_state(self):
        return self.admin

    @admin_state.setter
    def admin_state(self, state):
        """
        Set the admin state. This will power the interface on or off
        :param state: new power state, "up" to turn interface on, "down" to
            turn it off.
        """
        self.admin = state
        if (
            "lag" not in self.name
            and hasattr(self, "user_config")
            and isinstance(self.user_config, dict)
        ):
            self.user_config["admin"] = state

    def configure_vsx(
        self, active_forwarding, vsx_sync, act_gw_mac, act_gw_ip
    ):
        """
        Configure VSX IPv4 settings on a VLAN Interface.
        :param active_forwarding: True or False Boolean to set VSX active
            forwarding.
        :param vsx_sync: List of alphanumeric values to enable VSX
            configuration synchronization.  The options are
            any combination of 'active-gateways', 'irdp', and 'policies'. VSX
            Sync is mainly used in the Primary.
        :param act_gw_mac: Alphanumeric value of the Virtual MAC address for
            the interface active gateway. Example: '01:02:03:04:05:06'
        :param act_gw_ip: Alphanumeric value of the Virtual IP address for the
            interface active gateway. Example: '1.1.1.1'
        :return: True if object was changed.
        """
        # Set values
        vsx_sync_list = []
        if "active-gateways" in vsx_sync:
            vsx_sync_list.append("^vsx_virtual.*")
        if "irdp" in vsx_sync:
            vsx_sync_list.append(".irdp.*")
        if "policies" in vsx_sync:
            vsx_sync_list.append("^policy.*")

        self.vsx_active_forwarding_enable = active_forwarding
        self.vsx_sync = vsx_sync_list
        self.vsx_virtual_gw_mac_v4 = act_gw_mac
        self.vsx_virtual_ip4 = [act_gw_ip]

        # Apply changes
        return self.apply()

    def delete_vsx_configuration(self):
        """
        Delete VSX IPv4 settings on a VLAN Interface.
        :return: True if object was changed.
        """
        # Set values
        self.vsx_active_forwarding_enable = False
        self.vsx_sync = []
        self.vsx_virtual_gw_mac_v4 = None
        self.vsx_virtual_ip4 = []

        # Apply changes
        return self.apply()

    def configure_l3_ipv4_port(
        self, ip_address=None, port_desc=None, vrf="default"
    ):
        """
        Function will enable routing on the port and update the IPv4 address
            if given.
        :param ip_address: IPv4 address to assign to the interface. Defaults
            to nothing if not specified. Example: '1.1.1.1'
        :param port_desc: Optional description for the interface. Defaults to
            nothing if not specified.
        :param vrf: Name of the VRF to which the Port belongs. Defaults to
            "default" if not specified.
        :return: True if object was changed.
        """
        # Set IPv4

        if ip_address is not None:
            self.ip4_address = ip_address

        # Set description
        if port_desc is not None:
            self.description = port_desc

        # Set vrf
        vrf_obj = self.session.api.get_module(self.session, "Vrf", vrf)
        vrf_obj.get()
        self.vrf = vrf_obj

        # Set routing
        self.routing = True

        # Apply Changes inside Switch
        return self.apply()

    def update_ospf_interface_authentication(
        self, vrf, auth_type, digest_key, auth_pass
    ):
        """
        Perform PUT calls to update an Interface with OSPF to have
            authentication.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to.
        :param auth_type: Alphanumeric type of authentication, chosen between
            'md5', 'null', and 'text'.
        :param digest_key: Integer between 1-255 that functions as the digest
            key for the authentication method.
        :param auth_pass: Alphanumeric text for the authentication password.
            Note that this will be translated to a base64 String in the
            configuration and json.
        :return: True if object was changed.
        """
        # Configure Port/Interface
        self.configure_l3_ipv4_port(vrf=vrf)

        self.ospf_auth_type = auth_type
        self.ospf_auth_md5_keys = {str(digest_key): auth_pass}
        self.ospf_if_type = "ospf_iftype_broadcast"
        self.routing = True
        # Set vrf
        vrf_obj = self.session.api.get_module(self.session, "Vrf", vrf)
        vrf_obj.get()
        self.vrf = vrf_obj

        # Apply changes
        return self.apply()

    def update_ospf_interface_type(self, vrf, interface_type="pointtopoint"):
        """
        Update the Interface's OSPFv2 type, as well as enable routing on the
            interface.
        :param vrf: Alphanumeric name of the VRF the OSPF ID belongs to.
        :param interface_type: Alphanumeric type of OSPF interface.
            The options are 'broadcast', 'loopback', 'nbma', 'none',
            'pointomultipoint', 'pointopoint', and 'virtuallink'. Defaults to
            'pointtopoint'.
        :return: True if object was changed.
        """
        _valid_interface_types = [
            "broadcast",
            "loopback",
            "statistics",
            "nbma",
            "pointomultipoint",
            "pointopoint",
            "virtuallink",
            "none",
        ]
        if interface_type not in _valid_interface_types:
            raise Exception(
                "ERROR: interface_type must be one of: {0}".format(
                    _valid_interface_types
                )
            )

        # Configure Port/Interface
        self.configure_l3_ipv4_port(vrf=vrf)

        self.ospf_if_type = "ospf_iftype_{0}".format(interface_type)
        self.routing = True
        # Set vrf
        vrf_obj = self.session.api.get_module(self.session, "Vrf", vrf)
        vrf_obj.get()
        self.vrf = vrf_obj

        # Apply changes
        return self.apply()

    def set_active_gateway(self, ip_address, gateway_mac):
        """
        Update Active Gateway of a Interface.
        :param ip_address: IPv4 address to assign to the interface. Example:
            '1.1.1.1'.
        :param gateway_mac: Active Gateway MAC address to assign to the
            interface. Example: '01:02:03:04:05:06'.
        :return: True if object was changed.
        """
        # Configure Active Gateaway IP
        self.vsx_virtual_ip4 = [ip_address]
        # Configure Gateaway mac
        self.vsx_virtual_gw_mac_v4 = gateway_mac

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def update_interface_qos(self, qos):
        """
        Update QoS attached to this Interface.
        :param qos: string to define a QoS to operate on this interface. Use
            None to remove the Qos attached to this interface.
        :return: True if object was changed.
        """
        # Verify argument type and value
        if not isinstance(qos, str) and qos is not None:
            raise ParameterError("ERROR: QoS must be in a string format")

        self.qos = qos

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def update_interface_queue_profile(self, queue_profile):
        """
        Update the Queue Profile for this interface.
        :param queue_profile: Queue Profile name for this Interface.
            None is used to remove an existing Queue Profile.
        :return: True if object was changed.
        """
        if queue_profile is not None and not isinstance(queue_profile, str):
            raise ParameterError(
                "ERROR: queue_profile must be a string or None"
            )
        self.q_profile = queue_profile

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def update_interface_qos_trust_mode(
        self, qos_trust_mode, cos_override=None, dscp_override=None
    ):
        """
        Update the QoS trust mode of this port.
        :param qos_trust_mode: string to define the QoS trust mode for the
            interface. It can be either "cos", "dscp" or "none". To set the
            interface to use the global configuration use "global" instead.
        :param cos_override: integer with the COS entry id to associate
            with the interface instead of automatic values. In range [0,7]
        :param dscp_override: integer with the DSCP entry id to associate
            with the interface instead of the automatic values. In the
            range [0,63].
        :return: True if object was changed.
        """
        # Verify argument type and value
        if not isinstance(qos_trust_mode, str):
            raise ParameterError(
                "ERROR: QoS trust mode must be in a string format"
            )

        allowed_trust_modes = ["cos", "dscp", "none", "global"]
        if qos_trust_mode not in allowed_trust_modes:
            raise VerificationError(
                "ERROR: QoS trust mode must be one of: ", allowed_trust_modes
            )

        # Set trust mode in a key-value format
        if qos_trust_mode == "global":
            if "qos_trust" in self.qos_config:
                del self.qos_config["qos_trust"]
        else:
            self.qos_config["qos_trust"] = qos_trust_mode

        if cos_override:
            if not isinstance(cos_override, int):
                raise ParameterError(
                    "ERROR: COS Override must be in integer" "format"
                )
            self.qos_config["cos_override"] = cos_override
        if dscp_override:
            if not isinstance(dscp_override, int):
                raise ParameterError(
                    "ERROR: DSCP Override must be in integer" "format"
                )
            self.cos_config["dscp_override"] = dscp_override

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def update_interface_qos_rate(self, qos_rate):
        """
        Update the rate limit values configured for
            broadcast/multicast/unknown unicast traffic.
        :param qos_rate: dict of the rate limit values; should have the
            format ['<type of traffic>'] = <value><unit> e.g.
            {
                'unknown-unicast': '100pps',
                'broadcast': 200pps,
                'multicast': '200pps'
            }.
        :return: True if object was changed.
        """
        rate_limits = {}
        if qos_rate is not None:
            for traffic_type, rate in qos_rate.items():
                for idx, char in enumerate(rate):
                    if not char.isdigit():
                        break
                number = rate[:idx]
                unit = rate[idx:]

                rate_limits[traffic_type] = int(number)
                rate_limits[traffic_type + "_units"] = unit

        self.rate_limits = rate_limits

        # Apply changes
        return self.apply()

    def update_acl_in(self, acl_name, list_type):
        """
        Perform GET and PUT calls to apply ACL on an interface. This function
            specifically applies an ACL to Ingress traffic of the interface.
        :param acl_name: Alphanumeric String that is the name of the ACL
        :param list_type: Alphanumeric String of IPv4, IPv6 or MAC to specify
            the type of ACL.
        :return: True if object was changed.
        """
        # Create Acl object
        acl_obj = self.session.api.get_module(
            self.session, "ACL", index_id=acl_name, list_type=list_type
        )
        # Get the current version
        acl_obj.get()

        if list_type == "ipv6":
            self.aclv6_in_cfg = acl_obj
            if (
                hasattr(self, "aclv6_in_cfg_version")
                and self.aclv6_in_cfg_version is None
            ):
                self.aclv6_in_cfg_version = acl_obj.cfg_version
        elif list_type == "ipv4":
            self.aclv4_in_cfg = acl_obj
            if (
                hasattr(self, "aclv4_in_cfg_version")
                and self.aclv4_in_cfg_version is None
            ):
                self.aclv4_in_cfg_version = acl_obj.cfg_version
        elif list_type == "mac":
            self.aclmac_in_cfg = acl_obj
            if (
                hasattr(self, "aclmac_in_cfg_version")
                and self.aclmac_in_cfg_version is None
            ):
                self.aclmac_in_cfg_version = acl_obj.cfg_version

        # Apply changes
        return self.apply()

    def update_acl_out(self, acl_name, list_type):
        """
        Perform GET and PUT calls to apply ACL on an interface. This function
            specifically applies an ACL to Egress traffic of the interface,
            which must be a routing interface.
        :param acl_name: Alphanumeric String that is the name of the ACL.
        :param list_type: Alphanumeric String of IPv4, IPv6 or MAC to specify
            the type of ACL.
        :return: True if object was changed.
        """
        # Create Acl object
        acl_obj = self.session.api.get_module(
            self.session, "ACL", index_id=acl_name, list_type=list_type
        )
        # Get the current version
        acl_obj.get()

        if list_type == "ipv6":
            self.aclv6_out_cfg = acl_obj
            if (
                hasattr(self, "aclv6_out_cfg_version")
                and self.aclv6_out_cfg_version is None
            ):
                self.aclv6_out_cfg_version = acl_obj.cfg_version
        elif list_type == "ipv4":
            self.aclv4_out_cfg = acl_obj
            if (
                hasattr(self, "aclv4_out_cfg_version")
                and self.aclv4_out_cfg_version is None
            ):
                self.aclv4_out_cfg_version = acl_obj.cfg_version
        elif list_type == "mac":
            self.aclmac_out_cfg = acl_obj
            if (
                hasattr(self, "aclmac_out_cfg_version")
                and self.aclmac_out_cfg_version is None
            ):
                self.aclmac_out_cfg_version = acl_obj.cfg_version

        # Routeing
        self.routing = True

        # Apply changes
        return self.apply()

    @PyaoscxModule.materialized
    def port_security_enable(
        self,
        client_limit=None,
        sticky_mac_learning=None,
        allowed_mac_addr=None,
        allowed_sticky_mac_addr=None,
        violation_action=None,
        violation_recovery_time=None,
        violation_shutdown_recovery_enable=None,
    ):
        """
        Enable port security on an specified Interface.
        :param client_limit: Integer with the maximum amount of MAC
            addresses that can connect to the port.
        :param sticky_mac_learning: Boolean, If sticky MAC learning
            should be enabled.
        :param allowed_mac_addr: The list of allowed MAC addresses,
            each MAC address is a string, or a netaddr.EUI object.
        :param allowed_sticky_mac_addr: A dictionary where each key is
            a MAC address (string or netaddr.EUI), and the value is a list of
            integers, where each integer is a VLAN id.
        :param violation action: Action to take when unauthorized MACs
            try to connect.
        :param violation_recovery_time: integer with Time in seconds
            to wait for recovery.
        :param violation_shutdown_recovery_enable: Enable recovering
            from violation shut down.
        :return: True if the object was changed.
        """
        if not hasattr(self, "port_security"):
            raise VerificationError(
                "Unable to configure the port's security",
                "Interface {0} is not security capable".format(self.name),
            )

        if hasattr(self, "routing") and self.routing:
            raise VerificationError(
                "Configuring port-security is allowed only on bridged ports."
            )

        self.port_security["enable"] = True

        if client_limit is not None and (
            1 > client_limit or client_limit > 64
        ):
            raise ParameterError("Can only authorize 1 to 64 clients")

        if client_limit:
            self.port_security["client_limit"] = client_limit

        if sticky_mac_learning:
            self.port_security[
                "sticky_mac_learning_enable"
            ] = sticky_mac_learning

        # Use netaddr.EUI to verify that all static MAC addresses are valid,
        # and use a colon (:) character as the separator.  Example: accept
        # '2C:54:91:88:C9:E3', but not '2C-54-91-88-C9-E3', then re-convert to
        # string because that is what the API accepts.
        mac_format = mac_eui48
        mac_format.word_sep = ":"
        _valid_static_macs = []
        allowed_mac_addr = allowed_mac_addr or []
        for mac_addr in allowed_mac_addr:
            try:
                mac = MacAddress(mac_addr, dialect=mac_format)
                _valid_static_macs.append(str(mac))
            except AddrFormatError as exc:
                raise ParameterError("Invalid static MAC address") from exc
        self.port_security_static_client_mac_addr = _valid_static_macs

        # Use netaddr.EUI to verify that all static MAC addresses are valid,
        # and use a colon (:) character as the separator.  Example: accept
        # '2C:54:91:88:C9:E3', but not '2C-54-91-88-C9-E3', then re-convert to
        # string because that is what the API accepts.
        _valid_sticky_macs = {}
        allowed_sticky_mac_addr = allowed_sticky_mac_addr or {}
        for mac_address, vlans in allowed_sticky_mac_addr.items():
            try:
                mac = MacAddress(mac_address, dialect=mac_format)
                _valid_sticky_macs[str(mac)] = vlans
            except AddrFormatError as exc:
                raise ParameterError("Invalid sticky MAC address") from exc

        # Verify all VLAN IDs are valid numbers
        for mac_address, vlans in _valid_sticky_macs.items():
            _valid_vlans = []
            for vlan in vlans:
                try:
                    vlan = int(vlan)
                except ValueError as exc:
                    raise ParameterError("Invalid sticky MAC VLANs") from exc
                _valid_vlans.append(vlan)

            __status_int = Interface(self.session, self.name)
            __status_int.get(selector="status")
            # NOTE: applied_vlan_tag is NOT in the default get() path, so we
            # get it as a dictionary here
            _vlan_tag = None
            _vlan_tag_present = bool(__status_int.applied_vlan_tag)
            if _vlan_tag_present:
                _vlan_tag = int(next(iter(__status_int.applied_vlan_tag)))
            # NOTE: applied_vlan_trunks is NOT in the default get() path, so we
            # get it as a dictionary here
            _vlan_trunks = None
            _vlan_trunks_present = bool(__status_int.applied_vlan_trunks)
            if _vlan_trunks_present:
                _vlan_trunks = sorted(
                    [int(k) for k in __status_int.applied_vlan_trunks]
                )
            if _valid_vlans == [] and _vlan_tag_present:
                _valid_vlans = [_vlan_tag]

            if not _vlan_tag_present and not _vlan_trunks_present:
                raise VerificationError(
                    "No VLANs are configured in this interface"
                )
            for vlan in _valid_vlans:
                if not (
                    _vlan_tag_present
                    and vlan == _vlan_tag
                    or _vlan_trunks_present
                    and vlan in _vlan_trunks
                ):
                    _err_msg_allowed_vlans = []
                    if _vlan_tag_present:
                        _err_msg_allowed_vlans.append(
                            "vlan_access: {0}".format(_vlan_tag)
                        )
                    if _vlan_trunks_present:
                        _err_msg_allowed_vlans.append(
                            "vlan_trunks: {0}".format(_vlan_trunks)
                        )
                    raise VerificationError(
                        "One or more of {0} VLANs are not configured, the "
                        "allowed VLANs for this interface are: {1}".format(
                            _valid_vlans, ", ".join(_err_msg_allowed_vlans)
                        )
                    )
                _valid_sticky_macs[mac_address] = _valid_vlans
        self.port_security_static_sticky_client_mac_addr = _valid_sticky_macs

        if violation_action:
            if violation_action == "notify" and violation_recovery_time:
                raise ParameterError(
                    "Must not specify recovery time when violation action is "
                    "'notify'"
                )
            self.port_access_security_violation["action"] = violation_action

        if violation_recovery_time:
            if 10 > violation_recovery_time or violation_recovery_time > 600:
                raise ParameterError(
                    "violation_recovery_time must be between 10s and 600s"
                )
            self.port_access_security_violation[
                "recovery_timer"
            ] = violation_recovery_time

        if violation_shutdown_recovery_enable:
            self.port_access_security_violation[
                "shutdown_recovery_enable"
            ] = violation_shutdown_recovery_enable

        return self.apply()

    def port_security_disable(self):
        """
        Disable port security on the specified interface.
        :return: True if the object was changed.
        """
        if not self.materialized:
            raise VerificationError(
                "interface {0}".format(self.name), "Object not materialized"
            )

        if not hasattr(self, "port_security"):
            # This interface is not security capable
            warn(
                "Interface {0} is not security capable".format(self.name),
                RuntimeWarning,
            )

        self.port_security["enable"] = False
        return self.apply()

    @PyaoscxModule.materialized
    def speed_duplex_configure(
        self, speeds=["1000"], duplex="half", autonegotiation="off"
    ):
        """
        Configure the Interface speed and duplex mode.
        :param speed: List of allowed Interface speeds.
        :param duplex_enable: "full" for full duplex or "half" for half duplex.
        :param autonegotiation: switch autonegotiation "on" or "off".
        :return: True if object changed.
        """
        self.user_config["autoneg"] = autonegotiation
        self.user_config["duplex"] = duplex
        self.user_config["speeds"] = ",".join(speeds)

        return self.apply()



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

import pyaoscx.utils.util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class DhcpRelay(PyaoscxModule):
    """
    Provide configuration management for DHCP Relay on AOS-CX devices.
    """

    base_uri = "system/dhcp_relays"
    resource_uri_name = "dhcp_relays"

    indices = ["vrf", "port"]

    def __init__(self, session, vrf, port, uri=None, **kwargs):
        self.session = session
        # Assign IDs
        self.vrf = vrf
        self.port = port
        self._uri = uri
        # List used to determine attributes related to the DHCP Relay
        # configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a DHCP Relay table entry and
            fill the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}{2}{3}".format(
            DhcpRelay.base_uri,
            self.vrf.name,
            self.session.api.compound_index_separator,
            self.port.percents_name,
        )
        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Remove fields because they are not needed for the PUT request
        if "vrf" in data:
            data.pop("vrf")
        if "port" in data:
            data.pop("port")

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the DHCP Relay is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", ["vrf", "port"])

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "vrf" in self.__original_attributes:
            self.__original_attributes.pop("vrf")
        # Remove ID
        if "port" in self.__original_attributes:
            self.__original_attributes.pop("port")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system DHCP Relays, and create a
            dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return: Dictionary containing DHCP Relays IDs as keys and a DHCP Relay
            objects as values.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", DhcpRelay.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        dhcp_relay_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a DHCP Relay object
            indices, dhcp_relay = DhcpRelay.from_uri(session, uri)
            dhcp_relay_dict[indices] = dhcp_relay

        return dhcp_relay_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing DHCP Relay.
            Checks whether the DHCP Relay exists in the switch. Calls
            self.update() if DHCP Relay is being updated. Calls self.create()
            if a new DHCP Relay is being created.
        :return modified: Boolean, True if object was created or modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing DHCP Relay.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        dhcp_relay_data = utils.get_attrs(self, self.config_attrs)

        uri = "{0}/{1}{2}{3}".format(
            DhcpRelay.base_uri,
            self.vrf.name,
            self.session.api.compound_index_separator,
            self.port.percents_name,
        )

        # Compare dictionaries
        if dhcp_relay_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:

            post_data = json.dumps(dhcp_relay_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)
            # Set new original attributes
            self.__original_attributes = dhcp_relay_data

            # Object was modified, returns True
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new DHCP Relay. Only returns if no
            exception is raised.
        :return modified: Boolean, True if entry was created.
        """
        dhcp_relay_data = utils.get_attrs(self, self.config_attrs)
        dhcp_relay_data["vrf"] = self.vrf.get_info_format()
        dhcp_relay_data["port"] = self.port.get_info_format()

        post_data = json.dumps(dhcp_relay_data)

        try:
            response = self.session.request(
                "POST", DhcpRelay.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()

        # Object was created, means modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete DhcpRelay table entry.
        """
        uri = "{0}/{1}{2}{3}".format(
            DhcpRelay.base_uri,
            self.vrf.name,
            self.session.api.compound_index_separator,
            self.port.percents_name,
        )

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a DhcpRelay object given a response_data.
        :param cls: Class calling the method.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                {vrf},{port}: "/rest/v10.04/system/dhcp_relays/{vrf},{port}"
            }
        return: DhcpRelay object.
        """
        dhcp_relay_arr = session.api.get_keys(
            response_data, DhcpRelay.resource_uri_name
        )
        port_name = dhcp_relay_arr[1]
        vrf_name = dhcp_relay_arr[0]
        # Create Modules
        port_obj = session.api.get_module(session, "Interface", port_name)
        vrf_obj = session.api.get_module(session, "Vrf", vrf_name)

        return DhcpRelay(session, vrf_obj, port_obj)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a DHCP Relay object given a URI.
        :param cls: Class calling the method.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: a String with a URI.
        :return indices, dhcp_relay: tuple containing both the indices and
            DhcpRelay object.
        """
        # Obtain ID from URI
        index_pattern = re.compile(
            r"(.*)dhcp_relays/(?P<index1>.+)/(?P<index2>.+)"
        )
        vrf = index_pattern.match(uri).group("index1")
        port = index_pattern.match(uri).group("index2")

        port_obj = session.api.get_module(session, "Interface", port)
        vrf_obj = session.api.get_module(session, "Vrf", vrf)

        # Create DHCP Relay object
        dhcp_relay = DhcpRelay(session, vrf_obj, port_obj)
        indices = "{0},{1}".format(vrf, port)

        return indices, dhcp_relay

    def __str__(self):
        return "DhcpRelay vrf:{0}, port:{1}".format(self.vrf, self.port.name)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific DhcpRelay URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}{3}{4}".format(
                self.session.resource_prefix,
                DhcpRelay.base_uri,
                self.vrf.name,
                self.session.api.compound_index_separator,
                self.port.percents_name,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def add_ipv4_addresses(self, ipv4_list):
        """
        Perform a PUT calls to modify an existing DhcpRelay. Adding a list of
            IPv4 addresses into IPv4_ucast_server
        :param ipv4_list: List of IPv4 addresses. Example: ['1.1.1.1',
            '2.2.2.2']
        :return: True if object was changed.
        """
        # Set IPv4
        if ipv4_list is not None and ipv4_list != []:
            for i in range(len(ipv4_list)):
                if ipv4_list[i] not in self.ipv4_ucast_server:
                    self.ipv4_ucast_server.append(ipv4_list[i])

        # Apply changes inside switch
        return self.apply()

    def add_ipv6_addresses(self, ipv6_list):
        """
        Perform a PUT calls to modify an existing DhcpRelay. Adding a list of
            IPv6 addresses into IPv6_ucast_server.
        :param ipv6_list: List of IPv6 addresses. Example:
            ['2001:db8::11/ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff']
        :return: True if object was changed.
        """
        # Set IPv6
        if ipv6_list is not None and ipv6_list != []:
            for i in range(len(ipv6_list)):
                if ipv6_list[i] not in self.ipv6_ucast_server:
                    self.ipv6_ucast_server.append(ipv6_list[i])

        # Apply changes inside switch
        return self.apply()



# (C) Copyright 2021-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.unsupported_capability_error import (
    UnsupportedCapabilityError,
)

from pyaoscx.utils import util as utils

from pyaoscx.device import Device

from pyaoscx.pyaoscx_module import PyaoscxModule


class QosDscp(PyaoscxModule):
    """
    Provide configuration management for QoS DSCP trust mode on AOS-CX devices.
    """

    base_uri = "system/qos_dscp_map_entries"
    resource_uri_name = "qos_dscp_map_entries"

    indices = ["code_point"]

    def __init__(self, session, code_point, **kwargs):
        """
        Initialize a QoS DSCP trust mode object.
        :param session: pyaoscx.Session object used to represent logical
            connection to the device.
        :param code_point: Integer to identify a QoS DSCP configuration.
        """
        self.session = session
        self.__code_point = code_point

        # List used to determine attributes related to the QoS DSCP
        # configuration
        self.config_attrs = []
        self.materialized = False
        if "cos" in kwargs:
            self.cos = kwargs.pop("cos")
        if "priority_code_point" in kwargs:
            self.priority_code_point = kwargs.pop("priority_code_point")
        utils.set_creation_attrs(self, **kwargs)
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self._original_attributes = {}
        # Attribute used to know if object was changed recently
        self.__modified = False
        # Build path
        self.path = "{0}/{1}".format(self.base_uri, self.code_point)

    @property
    def code_point(self):
        """
        Method used to obtain the specific code point.
        :return: returns the code point of this QoS DSCP trust mode object.
        """
        # Use the @property decorator to make `self.code_point` read-only, and
        # return the actual value here
        return self.__code_point

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a QoS DSCP table entry and fill
            the object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to query specific information.
        :return: Returns True if there is not an exception raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        data = self._get_data(depth, selector)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        if selector in self.session.api.configurable_selectors:
            utils.set_config_attrs(self, data, "config_attrs")

        # Set original attributes
        self._original_attributes = data

        if "cos" in data:
            self.__cos = data["cos"]
        if "color" in data:
            self.__color = data["color"]
        if "description" in data:
            self.__description = data["description"]
        if "local_priority" in data:
            self.__local_priority = data["local_priority"]
        if "priority_code_point" in data:
            self.__priority_code_point = data["priority_code_point"]

        self.materialized = True

        return True

    @classmethod
    def get_all(cls, session):
        """
        Perform a GET call to retrieve all system QoS DSCP configurations from
            a switch.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :return:  containing all system QoS.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        try:
            response = session.request("GET", cls.base_uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        qos_dscp_dict = {}

        data = json.loads(response.text)
        uri_list = session.api.get_uri_from_data(data)
        for uri in uri_list:
            code_point, qos_dscp = cls.from_uri(session, uri)
            qos_dscp_dict[code_point] = qos_dscp

        return qos_dscp_dict

    @PyaoscxModule.materialized
    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to update an existing QoS table entry. Checks whether
            the QoS DSCP entry exists in the switch. Calls self.update if
            object is being updated.
        :return modified: Boolean, True if object was modified.
        """
        self.__modified = self.update()
        return self.__modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT request to update an existing QoS DSCP object.
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        qos_dscp_data = utils.get_attrs(self, self.config_attrs)

        return self._put_data(qos_dscp_data)

    @PyaoscxModule.connected
    def create(self):
        # TODO: Remove once abstractmethod decorator is removed from parent
        pass

    @PyaoscxModule.connected
    def delete(self):
        # TODO: Remove once abstractmethod decorator is removed from parent
        pass

    @classmethod
    def from_response(cls, session, response_data):
        """
        Create a QoS DSCP trust mode object given a response_data related to
            the QoS DSCP trust mode object.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param response_data: The response must be a dictionary of the form:
            {
                "3" : "/rest/v10.08/system/qos_dscp_map_entries/3"
            }
        :return: QoS DSCP trust mode object.
        """
        code_points_arr = session.api.get_keys(
            response_data, cls.resource_uri_name
        )

        code_point = code_points_arr[0]

        return cls(session, code_point)

    @classmethod
    def from_uri(cls, session, uri):
        """
        Create a QoS DSCP object given a QoS DSCP trust mode URI.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param uri: s String with a URI.
        :return: returns identifier and object.
        """
        # Separate values from URI
        if cls.base_uri not in uri:
            raise ValueError("Expected valid QoS DSCP trust mode URI.")

        # Extract code point from URI
        code_point = uri.split("/")[-1]

        qos_dscp = cls(session, code_point)

        # Return identifier and object
        return code_point, qos_dscp

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific QoS DSCP trust mode URI.
        :return: Object's URI.
        """
        # Return self.path containing the object's URI
        return self.path

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method to check it object has been modified.
        :return: Boolean True if the object was recently modified.
        """
        return self.modified

    def __str__(self):
        return "QoS DSCP trust mode {0}".format(self.code_point)

    @property
    def cos(self):
        """
        Getter method for the cos property.
        """
        return self.__cos

    @cos.setter
    def cos(self, cos):
        """
        Updates the value of the cos of this QoS DSCP instance.
        :param cos: Priority Code Point (PCP) that will be assigned to any IP
            packet with the specified DSCP codepoint, if that packet's ingress
            port has an effective trust mode of trust dscp. The new PCP is used
            when the packet is transmitted out a port or trunk with a VLAN tag.
            If the key is not specified, then no remark will occur.
        """
        # Verify data type
        if not isinstance(cos, int):
            raise ValueError("The value of cos must be an integer.")
        if not Device(self.session).is_capable("qos_cos_based_queueing"):
            raise UnsupportedCapabilityError(
                "This device doesn't support cos-based queueing"
            )
        self.__cos = cos

    @property
    def priority_code_point(self):
        """
        Getter method for the priority_code_point property.
        """
        return self.__priority_code_point

    @priority_code_point.setter
    def priority_code_point(self, priority_code_point):
        """
        Updates the value of the priority_code_point of this QoS DSCP instance.
        :param priority_code_point: Priority Code Point (PCP) that will be
            assigned to any IP packet with the specified DSCP codepoint, if
            that packet's ingress port has an effective trust mode of trust
            dscp. The new PCP is used when the packet is transmitted out a port
            or trunk with a VLAN tag.  If the key is not specified, then no
            remark will occur.
        """
        # Verify data type
        if not isinstance(priority_code_point, int):
            raise ValueError(
                "The value of priority_code_point must be an integer."
            )
        if not Device(self.session).is_capable("qos_dscp_map_cos_override"):
            raise UnsupportedCapabilityError(
                "This device doesn't support DSCP Map Cos Override"
            )
        self.__priority_code_point = priority_code_point

    @property
    def color(self):
        """
        Getter method for the color property.
        """
        return self.__color

    @color.setter
    def color(self, color):
        """
        Updates the value of the color of this QoS DSCP instance.
        :param color: String to identify the color which may be used later in
            the pipeline in packet-drop decision points. Example: "green".
        """
        # Verify data type
        if not isinstance(color, str):
            raise ValueError("ERROR: Color value must be a string.")
        self.__color = color

    @property
    def description(self):
        """
        Getter method for the description property.
        """
        return self.__description

    @description.setter
    def description(self, description):
        """
        Updates the description of this QoS DSCP instance.
        :param description: String used for customer documentation.
        """
        # Verify data type
        if not isinstance(description, str):
            raise ValueError("ERROR: Description value must be a string.")
        self.__description = description

    @property
    def local_priority(self):
        """
        Getter method for the local_priority.
        """
        return self.__local_priority

    @local_priority.setter
    def local_priority(self, priority):
        """
        Updates the value of the local priority of this QoS DSCP instance.
        :param priority: Integer to represent an internal meta-data value that
            will be associated with the packet. This value will be used later
            to select the egress queue for the packet.
        """
        # Verify data type
        if not isinstance(priority, int):
            raise ValueError("ERROR: Priority must be an integer.")
        self.__local_priority = priority



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging

from urllib.parse import quote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError
from pyaoscx.exceptions.verification_error import VerificationError

from pyaoscx.utils import util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class Configuration:
    """
    Represents a Device's Configuration and all of its attributes.
    Keeping all configuration information.
    """

    base_uri = "system"

    def __init__(self, session):
        self.session = session
        # Used to set attributes
        self.config_attrs = []
        self.materialized = False
        # Attribute used to know if object was changed recently
        self.__modified = False

    @PyaoscxModule.connected
    def get(self):
        """
        Perform a GET call to retrieve system attributes.
        """
        logging.info("Retrieving the switch attributes and capabilities")
        depth = self.session.api.default_depth
        uri = "{0}?depth={1}".format(Configuration.base_uri, depth)

        try:
            response = self.session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Create class attributes using util.create_attrs
        utils.create_attrs(self, data)

        # Second GET request to obtain just the variables that are writable
        selector = self.session.api.default_selector
        payload = {"depth": depth, "selector": selector}
        try:
            response = self.session.request(
                "GET", Configuration.base_uri, params=payload
            )

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Configurable data
        config_data = json.loads(response.text)

        # Set self.config_attrs and delete ID from it
        utils.set_config_attrs(self, config_data, "config_attrs")

        # Set original attributes
        self.__original_attributes = config_data

        # Set device as materialized
        self.materialized = True

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to update System Attributes. Checks whether the System
            is materialized. Calls self.update() if the configuration is being
            updated.
        :return modified: Boolean, True if object was modified.
        """
        modified = False
        if self.materialized:
            modified = self.update()
        else:
            raise VerificationError("Device", "Not materialized")
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to a Device Configuration.
        :return modified: Boolean, True if object was modified.
        """
        system_data = utils.get_attrs(self, self.config_attrs)

        # Compare dictionaries
        if system_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:

            put_data = json.dumps(system_data)

            try:
                response = self.session.request(
                    "PUT", Configuration.base_uri, data=put_data
                )

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text,
                    response.status_code,
                    "UPDATE SYSTEM ATTRIBUTES",
                )

            logging.info("SUCCESS: Updating System Attributes")
            # Set new original attributes
            self.__original_attributes = system_data

            # Object was modified, returns True
            modified = True

        return modified

    ####################################################################
    # IMPERATIVE FUNCTIONS
    ####################################################################

    def get_full_config(self, config_name="running-config"):
        """
        Perform a GET request to obtain the device's full config.
        :param config_name: String with the local-config name wanted.
            Defaults to 'running-config'.
        :return config_data: Data containing the full configuration.
        """
        uri = "fullconfigs/{0}".format(config_name)
        try:
            response = self.session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # Configurable data
        config_data = json.loads(response.text)

        return config_data

    def tftp_switch_config_from_remote_location(
        self, config_file_location, config_name, vrf
    ):
        """
        TFTP switch config from TFTP server.
        :param config_file_location: TFTP server address and path for
            uploading configuration.
        :param config_name: Config file or checkpoint to be uploaded to.
            When using TFTP only running-config or startup-config can be used.
        :param vrf: VRF to be used to contact TFTP server, required if
            remote_output_file_tftp_path is provided.
        :return success: Return True if response is successful or False if
            it was not.
        """
        success = False
        uri = "fullconfigs/{0}?from={1}&vrf={2}".format(
            config_name, config_file_location, vrf
        )

        try:
            response = self.session.request("PUT", uri)

            success = True

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        return success

    def copy_switch_config_to_remote_location(
        self, config_name, config_type, destination, vrf
    ):
        """
        Copy TFTP switch config to TFTP server using a PUT request.
        :param config_name:  String with the config file or checkpoint to be
            downloaded. When using TFTP only 'running-config' or
            'startup-config' can be used.
        :param config_type: Configuration type to be downloaded, JSON or CLI
            version of the config. 'json' or 'cli'.
        :param destination: TFTP server address and path for copying
            configuration, must be reachable through provided vrf.
        :param vrf: VRF to be used to contact TFTP server.
        :return: True if completed.
        """
        uri = "fullconfigs/{0}?to={1}&type={2}&vrf={3}".format(
            config_name, destination, config_type, vrf
        )
        try:
            response = self.session.request("GET", uri)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        # If no errors, return True for completion
        return True

    def backup_configuration(
        self,
        config_name,
        output_file=None,
        vrf=None,
        config_type="json",
        remote_file_tftp_path=None,
    ):
        """
        Obtains the switch's full config in json format and saves it to
            a local file or a remote location over TFTP.
        :param config_name:  String with the config file or checkpoint to be
            downloaded. When using TFTP only 'running-config' or
            'startup-config' can be used.
        :param output_file: String with the File name and path for locally
            downloading configuration, only JSON version of configuration will
            be downloaded.
        :param vrf: VRF to be used to contact TFTP server.
        :param config_type: Configuration type to be downloaded, JSON or CLI
            version of the config. 'json' or 'cli'. Defaults to 'json'.
        :param remote_file_tftp_path: TFTP server address and path for
            copying off configuration, must be reachable through provided vrf.
        :return bool: True if success.
        """
        success = False

        if remote_file_tftp_path is not None:
            tftp_path = remote_file_tftp_path
            if vrf is None:
                raise VerificationError(
                    "Backup Config",
                    "VRF needs to be provided in order to TFTP "
                    "the configuration from the switch",
                )

            tftp_path_encoded = quote_plus(tftp_path)

            if config_name not in ("running-config", "startup-config"):
                raise VerificationError(
                    "Backup Config",
                    "Only running-config or "
                    "startup-config can be backed-up using TFTP",
                )
            success = self.copy_switch_config_to_remote_location(
                config_name, config_type, tftp_path_encoded, vrf
            )
        else:
            config_json = self.get_full_config()
            with open(output_file, "w") as to_file:
                formatted_file = json.dumps(config_json, indent=4)
                to_file.write(formatted_file)

            success = True

        # Return result
        return success

    def create_checkpoint(self, source_config, destination_config):
        """
        Perform a PUT request to create a new checkpoint or copy an existing
              checkpoint to AOS-CX switch config.
        :param source_config: Name of the source configuration from which
            checkpoint needs to be created or copied.
        :param destination_config: Name of the destination configuration or
            name of checkpoint.
        :return bool: True if success.
        """
        success = False

        uri = "fullconfigs/{0}?from={1}fullconfigs/{2}".format(
            self.session.resource_prefix, destination_config, source_config
        )

        try:
            response = self.session.request("PUT", uri)

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        success = True

        # Return result
        return success

    def setup_mgmt_nameservers_dns(self, primary=None, secondary=None):
        """
        Setup primary and secondary name servers on a mgmt interface.
        :param primary: Primary nameservers on mgmt interface, a IPv4 address.
            Example: "10.10.2.10".
        :param secondary: Secondary nameservers on mgmt interface,
            a IP address. Example: "10.10.2.10".
        :return modified: Return True if coinfig was modified.
        """
        if "mode" in self.mgmt_intf:
            mgmt_if_mode = self.mgmt_intf["mode"]
        else:
            mgmt_if_mode = "dhcp"

        if mgmt_if_mode != "static":
            message_part1 = "The management interface must have static"
            message_part2 = "IP to configure management interface name servers"
            raise Exception(message_part1 + " " + message_part2)

        if primary is not None:
            self.mgmt_intf["dns_server_1"] = primary
        elif secondary is not None:
            self.mgmt_intf["dns_server_2"] = secondary

        return self.apply()

    def delete_mgmt_nameservers_dns(self):
        """
        Delete primary and secondary name servers on a mgmt interface.
        :return modified: Return True if coinfig was modified.
        """
        if "dns_server_1" in self.mgmt_intf:
            self.mgmt_intf.pop("dns_server_1")

        if "dns_server_2" in self.mgmt_intf:
            self.mgmt_intf.pop("dns_server_2")

        return self.apply()

    def upload_switch_config(
        self,
        config_name=None,
        config_file=None,
        config_json=None,
        vrf=None,
        remote_file_tftp_path=None,
    ):
        """
        Uploads configuration from a configuration file.
        :param config_name:  String with the Config file or checkpoint to be
            uploaded to. When using TFTP only 'running-config' or
            'startup-config' can be used. Defaults to None.
        :param config_file: File name for locally downloading configuration,
            only JSON version of configuration will be downloaded. Defaults to
            None.
        :param config_json: String with the JSON file name and path for locally
            uploading configuration, only JSON version of configuration can be
            uploaded. Defaults to None.
        :param vrf: String for VRF to be used to contact TFTP server, required
            if remote_output_file_tftp_path is provided. Defaults to None.
        :param remote_file_tftp_path: String for TFTP server address and path
            for copying off configuration, must be reachable through provided
            vrf. Defaults to None.
        :return success: Return boolean True if response is successful.
        """
        success = False

        if remote_file_tftp_path is not None:
            if vrf is None:
                raise VerificationError(
                    "Upload Config",
                    "VRF needs to be provided in order to TFTP "
                    "the configuration onto the switch",
                )

            tftp_path_encoded = quote_plus(remote_file_tftp_path)

            if config_name not in ("running-config", "startup-config"):
                raise VerificationError(
                    "Upload Config",
                    "Only running-config or startup-config "
                    "can be uploaded using TFTP",
                )

            success = self.tftp_switch_config_from_remote_location(
                tftp_path_encoded, config_name, vrf
            )

        else:

            success = self.upload_switch_config_from_local(
                config_json, config_file, config_name
            )

        return success

    def upload_switch_config_from_local(
        self, config_json=None, config_file=None, config_name=None
    ):
        """
        Uploads configuration from a configuration file.
        :param config_name:  String with the Config file or checkpoint
            to be uploaded to. When using TFTP only 'running-config' or
            'startup-config' can be used. Defaults to None.
        :param config_file: String with the File name and path for
            locally downloading configuration, only JSON version of
            configuration will be downloaded. Defaults to None.
        :param config_json: String with the JSON file name and path for
            locally uploading configuration, only JSON version of
            configuration can be uploaded. Defaults to None.
        :return success: Return boolean True if response is successful.
        """
        success = False

        if config_json:
            with open(config_json) as json_file:
                config_json = json.load(json_file)

        if config_file:
            with open(config_file) as json_file:
                config_json = json.load(json_file)

        config_json = json.dumps(config_json)

        # Create URI from the session base url and the configuration name
        uri = "fullconfigs/{0}".format(config_name)
        try:
            # Put (REST) configuration file
            response = self.session.request("PUT", uri, data=config_json)

            success = True

        except Exception as e:
            raise ResponseError("PUT", e)

        if not utils._response_ok(response, "PUT"):
            raise GenericOperationError(response.text, response.status_code)

        return success



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

import json
import logging
import re

from urllib.parse import quote_plus, unquote_plus

from pyaoscx.exceptions.generic_op_error import GenericOperationError
from pyaoscx.exceptions.response_error import ResponseError

import pyaoscx.utils.util as utils

from pyaoscx.pyaoscx_module import PyaoscxModule


class Ipv6(PyaoscxModule):
    """
    Provide configuration management for IPv6 on AOS-CX devices.
    """

    indices = ["address"]
    resource_uri_name = "ip6_addresses"

    def __init__(self, session, address, parent_int, uri=None, **kwargs):

        self.session = session
        # Assign address
        self.__set_name(address)
        # Assign parent Interface
        self.__set_interface(parent_int)
        self._uri = uri
        # List used to determine attributes related to the IPv6 configuration
        self.config_attrs = []
        self.materialized = False
        # Attribute dictionary used to manage the original data
        # obtained from the GET
        self.__original_attributes = {}
        # Set arguments needed for correct creation
        utils.set_creation_attrs(self, **kwargs)
        # Attribute used to know if object was changed recently
        self.__modified = False

    def __set_name(self, address):
        """
        Set name attribute in the proper form for IPv6.
        """
        # Add attributes to class
        self.address = None
        self.reference_address = None

        if r"%2F" in address or r"%3A" in address:
            self.address = unquote_plus(address)
            self.reference_address = address
        else:
            self.address = address
            self.reference_address = quote_plus(self.address)

    def __set_interface(self, parent_int):
        """
        Set parent interface as an attribute for the Ipv6 object.
        :param parent_int a Interface object.
        """
        # Set Parent Interface
        self.__parent_int = parent_int
        # Set Name for URI purposes
        self.__parent_int_name = self.__parent_int.percents_name

        # Set URI
        self.base_uri = "{0}/{1}/ip6_addresses".format(
            self.__parent_int.base_uri,
            self.__parent_int_name,
        )

        # Add self to ip6_address list in parent Interface
        for ip6_address in self.__parent_int.ip6_addresses:
            if ip6_address.address == self.address:
                # Make list element point to current object
                ip6_address = self

        # Adds to parent list
        self.__parent_int.ip6_addresses.append(self)

    @PyaoscxModule.connected
    def get(self, depth=None, selector=None):
        """
        Perform a GET call to retrieve data for a IPv6 table entry and fill the
            object with the incoming attributes.
        :param depth: Integer deciding how many levels into the API JSON that
            references will be returned.
        :param selector: Alphanumeric option to select specific information to
            return.
        :return: Returns True if no exception is raised.
        """
        logging.info("Retrieving %s from switch", self)

        depth = depth or self.session.api.default_depth
        selector = selector or self.session.api.default_selector

        if not self.session.api.valid_depth(depth):
            depths = self.session.api.valid_depths
            raise Exception("ERROR: Depth should be {0}".format(depths))

        if selector not in self.session.api.valid_selectors:
            selectors = " ".join(self.session.api.valid_selectors)
            raise Exception(
                "ERROR: Selector should be one of {0}".format(selectors)
            )

        payload = {"depth": depth, "selector": selector}

        uri = "{0}/{1}".format(self.base_uri, self.reference_address)

        try:
            response = self.session.request("GET", uri, params=payload)

        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        # Add dictionary as attributes for the object
        utils.create_attrs(self, data)

        # Determines if the IPv6 is configurable
        if selector in self.session.api.configurable_selectors:
            # Set self.config_attrs and delete ID from it
            utils.set_config_attrs(self, data, "config_attrs", ["address"])

        # Set original attributes
        self.__original_attributes = data
        # Remove ID
        if "address" in self.__original_attributes:
            self.__original_attributes.pop("address")
        # Remove type
        if "type" in self.__original_attributes:
            self.__original_attributes.pop("type")
        # Remove origin
        if "origin" in self.__original_attributes:
            self.__original_attributes.pop("origin")

        # Sets object as materialized
        # Information is loaded from the Device
        self.materialized = True
        return True

    @classmethod
    def get_all(cls, session, parent_int):
        """
        Perform a GET call to retrieve all system IPv6 addresses inside an
            Interface, and create a dictionary containing them.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_int: parent Interface object where IPv6 is stored
        :return: Dictionary containing IPv6 IDs as keys and a Ipv6 object as
            value.
        """
        logging.info("Retrieving all %s data from switch", cls.__name__)

        uri = "{0}/{1}/ip6_addresses".format(
            parent_int.base_uri, parent_int.percents_name
        )

        try:
            response = session.request("GET", uri)
        except Exception as e:
            raise ResponseError("GET", e)

        if not utils._response_ok(response, "GET"):
            raise GenericOperationError(response.text, response.status_code)

        data = json.loads(response.text)

        ipv6_dict = {}
        # Get all URI elements in the form of a list
        uri_list = session.api.get_uri_from_data(data)

        for uri in uri_list:
            # Create a Ipv6 object
            address, ipv6 = Ipv6.from_uri(session, parent_int, uri)
            # Load all IPv6 data from within the Switch
            ipv6.get()
            ipv6_dict[address] = ipv6

        return ipv6_dict

    @PyaoscxModule.connected
    def apply(self):
        """
        Main method used to either create or update an existing IPv6. Checks
            whether the IPv6 exists in the switch. Calls self.update() if IPv6
            is being updated. Calls self.create() if a new IPv6 is being
            created.
        """
        if not self.__parent_int.materialized:
            if self.__parent_int.__is_special_type:
                # Verify if it's a LAG
                self.__parent_int.apply()
            else:
                self.__parent_int.get()

        modified = False
        if self.materialized:
            modified = self.update()
        else:
            modified = self.create()
        # Set internal attribute
        self.__modified = modified
        return modified

    @PyaoscxModule.connected
    def update(self):
        """
        Perform a PUT call to apply changes to an existing IPv6 table entry
        :return modified: True if Object was modified and a PUT request was
            made.
        """
        # Variable returned
        modified = False

        ip6_data = utils.get_attrs(self, self.config_attrs)

        # Delete Type
        if "type" in ip6_data:
            ip6_data.pop("type")
        if "origin" in ip6_data:
            ip6_data.pop("origin")

        uri = "{0}/{1}".format(self.base_uri, self.reference_address)
        # Compare dictionaries
        if ip6_data == self.__original_attributes:
            # Object was not modified
            modified = False

        else:

            post_data = json.dumps(ip6_data)

            try:
                response = self.session.request("PUT", uri, data=post_data)

            except Exception as e:
                raise ResponseError("PUT", e)

            if not utils._response_ok(response, "PUT"):
                raise GenericOperationError(
                    response.text, response.status_code
                )

            logging.info("SUCCESS: Updating %s", self)

            # Set new original attributes
            self.__original_attributes = ip6_data

            # Object was modified
            modified = True
        return modified

    @PyaoscxModule.connected
    def create(self):
        """
        Perform a POST call to create a new IPv6 using the object's attributes
            as POST body. Only returns if an exception is not raised.
        :return modified: Boolean, True if entry was created.
        """
        ipv6_data = utils.get_attrs(self, self.config_attrs)
        ipv6_data["address"] = self.address

        post_data = json.dumps(ipv6_data)

        try:
            response = self.session.request(
                "POST", self.base_uri, data=post_data
            )

        except Exception as e:
            raise ResponseError("POST", e)

        if not utils._response_ok(response, "POST"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Adding %s", self)

        # Get all object's data
        self.get()
        # Object was created, thus modified
        return True

    @PyaoscxModule.connected
    def delete(self):
        """
        Perform DELETE call to delete IPv6 address from interface on the
            switch.
        """
        uri = "{0}/{1}".format(self.base_uri, self.reference_address)

        try:
            response = self.session.request("DELETE", uri)

        except Exception as e:
            raise ResponseError("DELETE", e)

        if not utils._response_ok(response, "DELETE"):
            raise GenericOperationError(response.text, response.status_code)

        logging.info("SUCCESS: Deleting %s", self)

        # Delete back reference from VRF
        for ip6 in self.__parent_int.ip6_addresses:
            if ip6.address == self.address:
                self.__parent_int.ip6_addresses.remove(ip6)

        # Delete object attributes
        utils.delete_attrs(self, self.config_attrs)

    @classmethod
    def from_response(cls, session, parent_int, response_data):
        """
        Create a IPv6 object given a response_data related to the IP6 address.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_int: parent Interface object where IPv6 is stored
        :param response_data: The response must be a dictionary of the form:
            {
                <address>: "/rest/v10.04/interface/ip6_addresses/<address>"
            }
        :return: IPv6 object.
        """
        ipv6_arr = session.api.get_keys(response_data, Ipv6.resource_uri_name)
        address = ipv6_arr[0]
        return Ipv6(session, address, parent_int)

    @classmethod
    def from_uri(cls, session, parent_int, uri):
        """
        Create a Ipv6 object given a URI.
        :param cls: Object's class.
        :param session: pyaoscx.Session object used to represent a logical
            connection to the device.
        :param parent_int: Parent Interface class where IPv6 is stored.
        :param uri: a String with a URI.
        :return index, ipv6_obj: tuple containing both the Ipv6 Object and
            the ipv6's address.
        """
        # Obtain ID from URI
        index_pattern = re.compile(r"(.*)ip6_addresses/(?P<index>.+)")
        index = index_pattern.match(uri).group("index")

        # Create Ipv6 object
        ipv6_obj = Ipv6(session, index, parent_int, uri=uri)

        return index, ipv6_obj

    def __str__(self):
        return "IPv6 address {0}".format(self.address)

    @PyaoscxModule.deprecated
    def get_uri(self):
        """
        Method used to obtain the specific IPv6 URI.
        return: Object's URI.
        """
        if self._uri is None:
            self._uri = "{0}{1}/{2}".format(
                self.session.resource_prefix,
                self.base_uri,
                self.reference_address,
            )

        return self._uri

    @PyaoscxModule.deprecated
    def get_info_format(self):
        """
        Method used to obtain correct object format for referencing inside
            other objects.
        return: Object format depending on the API Version.
        """
        return self.session.api.get_index(self)

    @property
    def modified(self):
        """
        Return boolean with whether this object has been modified.
        """
        return self.__modified

    @PyaoscxModule.deprecated
    def was_modified(self):
        """
        Getter method for the __modified attribute
        :return: Boolean True if the object was recently modified.
        """
        return self.modified



# (C) Copyright 2019-2022 Hewlett Packard Enterprise Development LP.
# Apache License 2.0

from pyaoscx.exceptions.pyaoscx_error import PyaoscxError


class VerificationError(PyaoscxError):
    """
    PYAOSCX Verification Error Exception.
    """

    base_msg = "VERIFICATION ERROR"

    def __init__(self, *args):
        self.message = None
        self.module = None
        if args:
            self.module = args[0]
            if len(args) > 1:
                self.message = ", ".join(str(a) for a in args[1:])

    def __str__(self):
        msg_parts = [self.base_msg]
        if self.module:
            if self.message:
                msg_parts.append("{0} DETAIL".format(self.module))
                msg_parts.append(self.message)
            else:
                msg_parts.append(self.module)
        msg = ": ".join(msg_parts)
        return repr(msg)
